# -*- coding: utf-8 -*-

"""The flowchart is a visual representation of a flowchart
drawn on a Tk canvas. The nodes of the graph are shown as
ovals, rectangles, etc. with the edges indicated by arrows
from one node to another.

The outline of a  node has the following Tk tags: ::

    node=xxxxx type=outline

The title: ::

    node=xxxxx type=title

When the mouse is over the node, the anchor points are activated.
They have the following tags: ::

    node=xxxxx type=anchor anchor=<point>

When the mouse is over one of the active anchor points, it is
covered with a larger circle, with tags: ::

    node=xxxxx type=active_anchor anchor=<point>

Edges are indicated by directional arrows between nodes. The
arrows have the following tags: ::

    edge=xxxxx type=arrow

When the mouse if over an arrow, it is shown to be active by placing
two red squares on the base and head of the arrow: ::

    type=arrow_base arrow=<item> edge=xxxxx
    type=arrow_head arrow=<item> edge=xxxxx

Clicking on either of these allows dragging the head or tail to
another anchor point on the same or another node (but not on the
same node as the tail/head for head/tail!). If the arrow is dropped
anywhere else it just snaps back to its original place.
"""

import copy
import logging
import math
from PIL import ImageTk, Image
import pkg_resources
import pprint  # nopep8
import sys
import tkinter as tk
import tkinter.filedialog as tk_filedialog
import tkinter.ttk as ttk

import seamm

logger = logging.getLogger(__name__)


def grey(value):
    return 255 - (255 - value) * 0.1


class TkFlowchart(object):

    def __init__(
        self, master=None, flowchart=None, namespace='org.molssi.seamm.tk'
    ):
        '''Initialize a Flowchart object

        Keyword arguments:
        '''

        self.toplevel = None
        self.master = master
        self._flowchart = flowchart
        self.filename = None
        self._stack = []

        self.graph = seamm.Graph()

        # Setup the plugin handling
        self.plugin_manager = seamm.PluginManager(namespace)

        # Job handler
        self._job_handler = seamm.TkJobHandler()

        self.canvas_width = 500
        self.canvas_height = 500
        self.grid_x = 300  # Width of the columns for the step display
        self.grid_y = 70  # Height of the rows for the step display
        self.gap = 50
        self.halo = 8  # How many pixels to consider 'near'
        self.data = None
        self._x0 = None
        self._y0 = None
        self.selection = None
        self.active_nodes = []
        self.in_callback = False
        self.canvas_after_callback = None
        self.popup_menu = None

        # Create the panedwindow
        self.pw = tk.PanedWindow(self.master, orient=tk.HORIZONTAL)
        self.pw.pack(fill=tk.BOTH, expand=1)

        # On the left put the tree of nodes
        self.tree = ttk.Treeview(self.pw)
        self.pw.add(self.tree)
        for group in self.plugin_manager.groups():
            self.tree.insert("", "end", group, text=group)
            for plugin in self.plugin_manager.plugins(group):
                self.tree.insert(
                    group, "end", plugin, text=plugin, tags="node"
                )
        self.tree.tag_bind(
            "node", sequence="<ButtonPress-1>", callback=self.create_node
        )

        # and the main canvas with scrollbars next to the right
        self.canvas_frame = ttk.Frame(self.pw)

        self.canvas_frame.grid_rowconfigure(0, weight=1)
        self.canvas_frame.grid_columnconfigure(0, weight=1)

        self.xscrollbar = tk.Scrollbar(self.canvas_frame, orient=tk.HORIZONTAL)
        self.xscrollbar.grid(row=1, column=0, sticky=tk.EW)

        self.yscrollbar = tk.Scrollbar(self.canvas_frame)
        self.yscrollbar.grid(row=0, column=1, sticky=tk.NS)

        self.canvas = tk.Canvas(
            self.canvas_frame,
            width=self.canvas_width,
            height=self.canvas_height,
            xscrollcommand=self.xscrollbar.set,
            yscrollcommand=self.yscrollbar.set,
            scrollregion=(0, 0, 2000, 2000)
        )
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)
        self.xscrollbar.config(command=self.xview)
        self.yscrollbar.config(command=self.yview)

        self.pw.add(self.canvas_frame)

        # Set up scrolling on the canvas with the mouse scrollwheel or similar
        self.canvas.bind('<Enter>', self._bound_to_mousewheel)
        self.canvas.bind('<Leave>', self._unbound_to_mousewheel)

        # background image
        filepath = pkg_resources.resource_filename(__name__, 'data/SEAMM.png')
        logger.info(filepath)

        self.image = Image.open(filepath)
        self.prepared_image = Image.eval(self.image.convert("RGB"), grey)
        w, h = self.image.size
        r_w = self.canvas_width / w
        r_h = self.canvas_height / h
        factor = r_w if r_w < r_h else r_h
        w = int(factor * w)
        h = int(factor * h)
        self.working_image = self.prepared_image.resize((w, h))
        self.photo = ImageTk.PhotoImage(self.working_image)
        # self.background = self.canvas.create_image(
        #     self.canvas_width / 2,
        #     self.canvas_height / 2,
        #     image=self.photo,
        #     anchor='center')
        self.background = self.canvas.create_image(
            0, 0, image=self.photo, anchor='center'
        )

        # The gui partner for the start node...
        self.create_start_node()

        # Set up the bindings
        self.canvas.bind('<Configure>', self.canvas_configure)
        self.canvas.bind('<Motion>', self.mouse_motion)
        self.canvas.bind('<ButtonPress-1>', self.click)
        self.canvas.bind('<Double-ButtonPress-1>', self.double_click)
        if sys.platform.startswith('darwin'):
            self.canvas.bind('<ButtonPress-2>', self.right_click)
        else:
            self.canvas.bind('<ButtonPress-3>', self.right_click)

        logger.debug('Finished initializing tk_flowchart')

    def __iter__(self):
        return self.graph.__iter__()

    @property
    def flowchart(self):
        """The flowchart, which holds the nodes"""
        return self._flowchart

    @flowchart.setter
    def flowchart(self, value):
        self._flowchart = value

    @property
    def master(self):
        """The window that is our master"""
        return self._master

    @master.setter
    def master(self, value):
        if value:
            self.toplevel = value.winfo_toplevel()
        self._master = value

    def tag_exists(self, tag):
        """Check if the node with a given tag exists"""
        for node in self:
            if node.tag == tag:
                return True
        return False

    def get_node(self, tag):
        """Return the node with a given tag"""
        if isinstance(tag, int):
            tag = str(tag)
        for node in self:
            if str(node.uuid) == tag:
                return node
        return None

    def last_node(self, tk_node='1'):
        """Find the last node walking down the main execution path
        from the given node, which defaults to the start node"""

        logger.debug('Finding last node')
        # Handle loops!
        for node in self:
            node.node.visited = False
            logger.debug(
                '   reset visited {} {} = {}'.format(
                    node.node.visited, node.title, node
                )
            )

        return self.last_node_helper(tk_node)

    def last_node_helper(self, tk_node):
        """Helper routine to handle the recursion"""

        # get the node to start the traversal
        if isinstance(tk_node, str):
            tk_node = self.get_node(tk_node)

        logger.debug(
            '   last tk_node helper: {} {} = {}'.format(
                tk_node.node.visited, tk_node.title, tk_node
            )
        )

        tk_node.node.visited = True
        next_tk_node = None
        for edge in self.graph.edges(tk_node, direction='out'):
            if edge.edge_type == "execution":

                if edge.node2.node.visited:
                    logger.debug(
                        '\ttk_node {} {} has been visited'.format(
                            edge.node2.title, edge.node2
                        )
                    )
                    next_tk_node = edge.node2
                else:
                    logger.debug(
                        '\trecursing to tk_node {} {}'.format(
                            edge.node2.title, edge.node2
                        )
                    )
                    return self.last_node_helper(edge.node2)

        if next_tk_node is not None:
            tk_node = next_tk_node
            logger.debug(
                '\tchecking visited tk_node {} {} for new nodes'.format(
                    tk_node.title, tk_node
                )
            )
            if tk_node.node.extension == "Join":
                logger.debug('\t  tk_node is a join node, so look at next')
                for edge in self.graph.edges(tk_node, direction='out'):
                    if edge.edge_type == "execution":
                        logger.debug(
                            '\ttk_node after join node is {} {}'.format(
                                edge.node2.title, edge.node2
                            )
                        )
                        tk_node = edge.node2

            for edge in self.graph.edges(tk_node, direction='out'):
                if edge.edge_type == "execution":
                    if edge.node2.node.visited:
                        logger.debug(
                            '\tnode {} {} has been visited'.format(
                                edge.node2.title, edge.node2
                            )
                        )
                    else:
                        logger.debug(
                            '\trecursing to tk_node {} {}'.format(
                                edge.node2.title, edge.node2
                            )
                        )
                        return self.last_node_helper(edge.node2)

        logger.debug('\treturning {} {}'.format(tk_node.title, tk_node))
        return tk_node

    def add_edge(
        self, u, v, edge_type='execution', edge_subtype='next', **kwargs
    ):
        edge = self.graph.add_edge(
            u,
            v,
            edge_type,
            edge_subtype,
            edge_class=seamm.TkEdge,
            canvas=self.canvas,
            **kwargs
        )
        edge.draw()
        return edge

    def edges(self, node=None, direction='both'):
        return self.graph.edges(node, direction)

    def new_file(self, event=None):
        self.filename = None
        self.clear()

    def help(self, event=None):
        print("Help!!!!")

    def debug(self, event):
        print(event)

    def open_file(self, event=None):
        filename = tk_filedialog.askopenfilename(defaultextension='.flow')
        if filename == '':
            return
        self.open(filename)

    def open(self, filename):
        if isinstance(filename, list):
            filename = filename[0]

        self.flowchart.read(filename)
        self.from_flowchart()
        self.filename = filename

    def clear(self, all=False):
        """Clear our graphics"""
        for item in self.canvas.find_all():
            if item != self.background:
                self.canvas.delete(item)
        # and the graph
        self.graph.clear()

        # recreate the start node
        if not all:
            self.create_start_node()
            self.draw()

    def create_start_node(self):
        """Create the start node"""
        # Check if the start node exists
        start_node = self.flowchart.get_node('1')
        if start_node is None:
            start_node = seamm.StartNode()

        tk_start_node = seamm.TkStartNode(
            tk_flowchart=self,
            canvas=self.canvas,
            node=start_node,
            x=self.grid_x / 2,
            y=self.grid_y / 2
        )

        self.graph.add_node(tk_start_node)
        logger.debug(
            'Created start node {} at {}, {}'.format(
                tk_start_node, start_node.x, start_node.y
            )
        )
        return tk_start_node

    def save(self, event=None):
        if self.filename is None:
            self.save_file()
        else:
            self.update_flowchart()
            self.flowchart.write(self.filename)
            self.flowchart.clear()

    def save_file(self, event=None):
        filename = tk_filedialog.asksaveasfilename(defaultextension='.flow')
        if filename != '':
            # suffixes = pathlib.Path(filename).suffixes
            # if len(suffixes) == 0 or '.flow' not in suffixes:
            #     filename = filename + '.flow'
            self.filename = filename
            self.save()

    def about(self, text='In about'):
        print(text)

    def preferences(self):
        print('In preferences')

    def draw(self):
        for tk_node in self:
            tk_node.draw()

    def canvas_configure(self, event):
        """Redraw the background as the canvas changes size

        Only after the process is idle!
        """
        if self.canvas_after_callback is None:
            self.canvas_after_callback = self.canvas.after_idle(
                self.canvas_configure_doit
            )

    def canvas_configure_doit(self):
        """Redraw the background as the canvas changes size

        This keeps the background image as large as possible and
        centered in the flowchart canvas.
        """
        if self.canvas_after_callback is not None:
            self.canvas.after_cancel(self.canvas_after_callback)
        self.canvas_after_callback = None

        w, h = self.image.size
        cw = self.canvas.winfo_width()
        ch = self.canvas.winfo_height()

        r_w = cw / w
        r_h = ch / h
        factor = r_w if r_w < r_h else r_h
        w = int(factor * w)
        h = int(factor * h)

        self.canvas.itemconfigure(self.background, image=None)
        # self.canvas.coords(self.background, cw / 2, ch / 2)
        self.canvas.coords(self.background, 0, 0)
        del (self.working_image)
        self.working_image = self.prepared_image.resize((w, h))
        del (self.photo)
        self.photo = ImageTk.PhotoImage(self.working_image)
        # self.canvas.itemconfigure(
        #     self.background, image=self.photo, anchor='center')
        self.canvas.itemconfigure(
            self.background, image=self.photo, anchor='nw'
        )

    def click(self, event):
        """Handle a left-click on the canvas by finding out what the
        mouse is on/in/near and doing the appropriate thing, such as
        selecting to preparing to move the item.
        """

        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        items = self.canvas.find_closest(cx, cy, self.halo)
        self.selection = []
        for item in items:
            tags = self.get_tags(item)
            if 'type' in tags and \
               (tags['type'] == 'arrow_base' or tags['type'] == 'arrow_head'):
                arrow = int(tags['arrow'])
                xys = self.canvas.coords(item)
                x0 = xys[0]
                y0 = xys[1]
                x1 = xys[-2]
                y1 = xys[-1]
                self.data = self.get_tags(arrow)
                self.data['arrow'] = arrow
                self.data.update(self.get_tags(arrow))
                self.data['x0'] = x0
                self.data['y0'] = y0
                self.data['x1'] = x1
                self.data['y1'] = y1
                self._x0 = cx
                self._y0 = cy

                logger.debug('self.data for dragging arrow head or base')
                logger.debug('{}'.format(self.data))

                if tags['type'] == 'arrow_base':
                    self.data['arrow_base'] = item
                    self.data['arrow_head'] = \
                        self.canvas.find_withtag('type=arrow_head')
                    self.mouse_op = 'drag arrow base'
                    self.canvas.bind('<B1-Motion>', self.drag_arrow_base)
                    self.canvas.bind('<ButtonRelease-1>', self.drop_arrow_base)
                else:
                    self.data['arrow_base'] = \
                        self.canvas.find_withtag('type=arrow_base')
                    self.data['arrow_head'] = item
                    self.mouse_op = 'drag arrow head'
                    self.canvas.bind('<B1-Motion>', self.drag_arrow_head)
                    self.canvas.bind('<ButtonRelease-1>', self.drop_arrow_head)

            if 'node' in tags:
                node = tags['node']
                if tags['type'] == 'active_anchor':
                    # Connecting from an anchor to another node
                    x, y = node.anchor_point(tags['anchor'])
                    self.mouse_op = 'Connect'
                    arrow = self.canvas.create_line(
                        x, y, cx, cy, arrow=tk.LAST, tags='type=active_arrow'
                    )
                    self.data = (node, tags['anchor'], x, y, arrow)
                    self.canvas.bind('<B1-Motion>', self.drag_arrow)
                    self.canvas.bind('<ButtonRelease-1>', self.drop_arrow)
                else:
                    if node.is_inside(cx, cy, self.halo):
                        self.selection.append(node)
                        node.selected = True
                        self._x0 = cx
                        self._y0 = cy
                        self.mouse_op = 'Move'
                        self.canvas.bind('<B1-Motion>', self.move)
                        self.canvas.bind('<ButtonRelease-1>', self.end_move)
                    else:
                        node.selected = False

    def double_click(self, event):
        """Handle a double-click on the canvas by finding out what the
        mouse is on/in/near and doing the appropriate thing.
        """

        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        result = self.find_items(cx, cy)
        self.selection = []
        if result is None:
            # Handle a right-click out side anything
            print('Double-click outside objects')
            return

        if result[0] == 'node':
            node = result[1]
            if node.is_inside(cx, cy):
                node.double_click(event)
                return

        if result[0] == 'item':
            item = result[1]
            tags = self.get_tags(item)
            if 'type' in tags and tags['type'] == 'arrow':
                self.right_click_on_arrow(event, item, tags)

    def right_click(self, event):
        """Handle a right-click on the canvas by finding out what the
        mouse is on/in/near and doing the appropriate thing, such as
        posting an action menu
        """

        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        result = self.find_items(cx, cy)
        self.selection = []
        if result is None:
            # Handle a right-click out side anything
            print('Right-click outside objects')
            return

        if result[0] == 'node':
            node = result[1]
            if node.is_inside(cx, cy):
                node.right_click(event)
                return

        if result[0] == 'item':
            item = result[1]
            tags = self.get_tags(item)
            if 'type' in tags and tags['type'] == 'arrow':
                self.right_click_on_arrow(event, item, tags)

    def move(self, event):
        '''Move selected items
        '''

        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        deltax = cx - self._x0
        deltay = cy - self._y0

        self._x0 = cx
        self._y0 = cy

        for item in self.selection:
            item.move(deltax, deltay)

    def end_move(self, event):
        '''End the move of selected items
        '''
        self.canvas.bind('<B1-Motion>', '')
        self.canvas.bind('<ButtonRelease-1>', '')

        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        deltax = cx - self._x0
        deltay = cy - self._y0

        for item in self.selection:
            item.end_move(deltax, deltay)
            item.selected = False

        self._x0 = None
        self._y0 = None
        self.mouse_op = None
        self.selection = None

    def create_node(self, event):
        """Create a node using the type in menu. This is a bit tricky because
        we need to create both the node and its graphical partner, each of
        which needs to know the other.
        """

        # not scrolling the left pane yet
        item = self.tree.identify_row(event.y)
        plugin_name = self.tree.item(item, option="text")

        (last_node, x, y, anchor1, anchor2) = self.next_position()
        edge_subtype = last_node.default_edge_subtype()

        logger.debug('creating {} node'.format(plugin_name))

        # The node.
        node = self.flowchart.create_node(plugin_name)

        # The graphics partner
        plugin = self.plugin_manager.get(plugin_name)
        logger.debug('  plugin object: {}'.format(plugin))
        tk_node = plugin.create_tk_node(
            tk_flowchart=self, canvas=self.canvas, x=0, y=0, node=node
        )
        self.graph.add_node(tk_node)

        # And figure out where the node should be
        # Use the grid approach
        x0 = last_node.x
        y0 = last_node.y

        if anchor1 == 's':
            tk_node.x = x0
            tk_node.y = y0 + self.grid_y
        elif anchor1 == 'e':
            tk_node.x = x0 + self.grid_x
            tk_node.y = y0
        else:
            dx, dy = tk_node.anchor_point(anchor2)

            # flip sign to get direction right
            tk_node.x = x - dx
            tk_node.y = y - dy

        # And connect this to the last node in the existing flowchart,
        # which is probably what the user wants.

        self.add_edge(
            last_node,
            tk_node,
            'execution',
            anchor1=anchor1,
            anchor2=anchor2,
            edge_subtype=edge_subtype
        )

        # And update the picture on screen
        self.draw()

    def remove_node(self, node):
        """Remove the given node"""

        # remove the graphical rendering
        node.undraw()

        # and edges
        node.remove_edge('all')

        # and the node itself
        self.graph.remove_node(node)

    def next_position(self):
        """Find a reasonable place to position the next step
        in the flowchart."""

        last_node = self.last_node()

        # center of node
        x0 = last_node.x
        y0 = last_node.y

        # Get the anchor point the last node wants to use
        anchor1 = last_node.next_anchor()
        # and the inverse for the new node
        anchor2 = anchor1.translate(''.maketrans('news', 'swen'))

        # Find the point 'gap' past the anchor point of the last
        # node, looking from the center (0, 0)

        x1, y1 = last_node.anchor_point(anchor1)
        dx = x1 - x0
        dy = y1 - y0
        norm = math.sqrt(dx * dx + dy * dy)

        x = x1 + self.gap * (dx / norm)
        y = y1 + self.gap * (dy / norm)

        return (last_node, x, y, anchor1, anchor2)

    def mouse_motion(self, event, exclude=()):
        """Track the mouse and highlight the node under the mouse

        It appears that the canvas find_closest does not work properly in
        Python. Even if you give it a tag to look below, it always returns
        the topmost item, so we cannot loop through items.

        Instead we use find_overlapping, which does return a list. However,
        if the mouse is e.g. inside a rectangle but far enough from the edges
        find_overlapping does not find it. In this case we use the current
        tag to find the object.
        """

        if self.in_callback:
            print("IN CALLBACK!!!!!!")
            return
        self.in_callback = True

        result = None

        self.canvas.delete('type=active_anchor')
        self.canvas.delete('type=arrow_base')
        self.canvas.delete('type=arrow_head')

        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))

        active = []
        items = self.canvas.find_overlapping(
            cx + self.halo / 2, cy + self.halo / 2, cx - self.halo / 2,
            cy - self.halo / 2
        )
        if len(items) == 0:
            # If we are within e.g. a rectangle, it may not overlap
            # but will be the current item, so if nothing overlaps
            # use the current item (if there is one).

            items = self.canvas.find_withtag('current')

        # Loop backwards since the 'top' item is at the end of the list
        # and is probably the item we want.

        for item in items[::-1]:
            if item in exclude:
                continue
            tags = self.get_tags(item)

            # on an arrow?
            if 'type' in tags and tags['type'] == 'arrow':
                xys = self.canvas.coords(item)
                x0 = xys[0]
                y0 = xys[1]
                x1 = xys[-2]
                y1 = xys[-1]
                self.canvas.create_rectangle(
                    x0 - self.halo / 2,
                    y0 - self.halo / 2,
                    x0 + self.halo / 2,
                    y0 + self.halo / 2,
                    tags=[
                        'type=arrow_base', 'arrow=' + str(item),
                        tags['edge'].tag()
                    ],
                    outline='red',
                    fill='red'
                )
                self.canvas.create_rectangle(
                    x1 - self.halo / 2,
                    y1 - self.halo / 2,
                    x1 + self.halo / 2,
                    y1 + self.halo / 2,
                    tags=[
                        'type=arrow_head', 'arrow=' + str(item),
                        tags['edge'].tag()
                    ],
                    outline='red',
                    fill='red'
                )
                break
            if 'node' in tags:
                node = tags['node']
                if node.is_inside(cx, cy, self.halo):
                    active.append(node)
                    if node not in self.active_nodes:
                        node.activate()
                        self.active_nodes.append(node)
                    # are we close to any anchor points?
                    point = node.check_anchor_points(cx, cy, self.halo)
                    if point is None:
                        self.canvas.delete('type=active_anchor')
                    else:
                        node.activate_anchor_point(point, self.halo)
                        result = (node, point)
                    break

        # deactivate any previously active nodes
        for node in self.active_nodes:
            if node not in active:
                node.deactivate()
        self.active_nodes = active

        self.in_callback = False

        return result

    def find_items(self, x, y, exclude=()):
        """Return the 'top' node under the mouse coordinates x, y

        It appears that the canvas find_closest does not work properly in
        Python. Even if you give it a tag to look below, it always returns
        the topmost item, so we cannot loop through items.

        Instead we use find_overlapping, which does return a list. However,
        if the mouse is e.g. inside a rectangle bat far enough from the edges
        find_overlapping does not find it. In this case we use the current
        tag to find the object.
        """

        items = self.canvas.find_overlapping(
            x + self.halo / 2, y + self.halo / 2, x - self.halo / 2,
            y - self.halo / 2
        )
        if len(items) == 0:
            # If we are within e.g. a rectangle, it may not overlap
            # but will be the current item, so if nothing overlaps
            # use the current item (if there is one).

            items = self.canvas.find_withtag('current')

        # Loop backwards since the 'top' item is at the end of the list
        # and is probably the item we want.

        for item in items[::-1]:
            if item in exclude:
                continue

            tags = self.get_tags(item)
            if 'node' in tags:
                node = tags['node']
                if node.is_inside(x, y, self.halo):
                    # are we close to any anchor points?
                    point = node.check_anchor_points(x, y, self.halo)
                    return ('node', node, point)
            return ('item', item)
        return None

    def get_tags(self, item):
        '''Return the tags of "item" as a dict. Any added tags
        like "active" are added to the "extra" dict entry.
        '''

        tags = {}
        tags['extra'] = []
        for x in self.canvas.gettags(item):
            if '=' in x:
                key, value = x.split('=')
                if 'node' in key:
                    tags[key] = self.get_node(value)
                elif 'edge' == key:
                    tags[key] = seamm.TkEdge.str_to_object[value]
                else:
                    tags[key] = value
            else:
                tags['extra'].append(x)
        return tags

    def drag_arrow(self, event):
        '''Drag an arrow from the anchor on the node to the mouse
        Used when creating a new edge.
        '''

        logger.debug('drag arrow')
        node, anchor, x, y, arrow = self.data
        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        logger.debug(
            '  x = {}, y = {}, cx = {}, cy = {}, arrow = {}'.format(
                x, y, cx, cy, arrow
            )
        )
        self.canvas.coords(arrow, x, y, cx, cy)
        # Check for being near another nodes anchor point
        result = self.find_items(cx, cy, exclude=(arrow,))
        logger.debug('  result = {}'.format(result))
        if result is not None and result[0] == 'node':
            logger.debug('       node = {}'.format(node))
            logger.debug('  result[1] = {}'.format(result[1]))
            if node == result[1]:
                self.canvas.delete('type=active_anchor')
                logger.debug('  deactivate {}'.format(result[1]))
                result[1].deactivate()
            else:
                logger.debug(
                    '  activate_node {} {}'.format(result[1], result[2])
                )
                self.activate_node(result[1], result[2])

    def drop_arrow(self, event):
        '''The user has dropped a new arrow somewhere!
        If it is on another node, make the connection.
        If it is in empty space or on the original node,
        just cancel the operation.
        '''

        self.canvas.bind('<B1-Motion>', '')
        self.canvas.bind('<ButtonRelease-1>', '')

        node, anchor, x, y, arrow = self.data
        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        self.canvas.coords(arrow, x, y, cx, cy)
        # Check for being near another nodes anchor point
        result = self.find_items(cx, cy)
        self.canvas.delete(arrow)

        if result is not None and result[0] == 'node':
            other_node, point = result[1:]
            if node != other_node and point is not None:
                edge_subtype = node.default_edge_subtype()
                self.add_edge(
                    node,
                    other_node,
                    'execution',
                    anchor1=anchor,
                    anchor2=point,
                    edge_subtype=edge_subtype
                )

        self.data = None
        self.mouse_op = None

    def drag_arrow_base(self, event):
        '''Drag the base of an exisiting arrow
        '''

        # move arrow
        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        self.canvas.coords(
            self.data['arrow'], cx, cy, self.data['x1'], self.data['y1']
        )

        # move base icon
        deltax = cx - self._x0
        deltay = cy - self._y0

        self._x0 = cx
        self._y0 = cy

        self.canvas.move(self.data['arrow_base'], deltax, deltay)

        # move the label if there is one
        edge = self.data['edge']
        if edge.has_label:
            self.canvas.coords(
                edge.label_id,
                edge.label_position(cx, cy, self.data['x1'], self.data['y1'])
            )
            self.canvas.coords(
                edge.label_bg_id, self.canvas.bbox(edge.label_id)
            )

        # Check for being near another nodes anchor point
        result = self.find_items(
            cx, cy, exclude=(self.data['arrow'], self.data['arrow_base'])
        )

        if result is not None and result[0] == 'node':
            self.activate_node(
                result[1], result[2], exclude=(self.data['edge'].node2,)
            )

    def activate_node(self, node, point=None, exclude=()):
        '''Activate a node, i.e. display the anchor points,
        unless it is in the exclusion list. Also, if the
        anchor point is given, make it active.
        '''

        active = []
        if node in exclude:
            self.canvas.delete('type=active_anchor')
            node.deactivate()
        else:
            active.append(node)
            if node not in self.active_nodes:
                node.activate()
                self.active_nodes.append(node)
            if point is None:
                self.canvas.delete('type=active_anchor')
            else:
                node.activate_anchor_point(point, self.halo)

        # deactivate any previously active nodes
        for node in self.active_nodes:
            if node not in active:
                node.deactivate()
        self.active_nodes = active

    def drop_arrow_base(self, event):
        '''The user has dropped the arrow somewhere!
        If it is on another node, make the connection.
        If it is in empty space or on the original node,
        just cancel the operation.
        '''

        self.canvas.bind('<B1-Motion>', '')
        self.canvas.bind('<ButtonRelease-1>', '')

        # Check for being near another nodes anchor point
        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        result = self.find_items(cx, cy)

        edge = self.data['edge']

        if result is None:
            # dropped on empty space
            self.canvas.delete('type=arrow_base')
            self.canvas.delete('type=arrow_head')
            edge.draw()
        elif result[0] == 'node':
            # dropped on another node
            node1, anchor1 = result[1:]
            if edge.node1 == node1:
                edge.anchor1 = anchor1
                edge.draw()
            else:
                # remove current connection and create new, in
                # that order -- otherwise tend to remove edge
                # completely if it is moved on same node.

                node2 = edge.node2
                anchor2 = edge.anchor2
                edge_subtype = edge.edge_subtype

                self.remove_edge(self.data['arrow'])

                self.add_edge(
                    node1,
                    node2,
                    edge_type='execution',
                    edge_subtype=edge_subtype,
                    anchor1=anchor1,
                    anchor2=anchor2,
                )

        self.data = None
        self.mouse_op = None

    def drag_arrow_head(self, event):
        '''Drag the head of an arrow
        '''

        # move arrow
        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        self.canvas.coords(
            self.data['arrow'], self.data['x0'], self.data['y0'], cx, cy
        )

        # move base icon
        deltax = cx - self._x0
        deltay = cy - self._y0

        self._x0 = cx
        self._y0 = cy

        self.canvas.move(self.data['arrow_head'], deltax, deltay)

        # move the label if there is one
        edge = self.data['edge']
        if edge.has_label:
            self.canvas.coords(
                edge.label_id,
                edge.label_position(self.data['x0'], self.data['y0'], cx, cy)
            )
            self.canvas.coords(
                edge.label_bg_id, self.canvas.bbox(edge.label_id)
            )

        # Check for being near another nodes anchor point
        result = self.find_items(
            cx, cy, exclude=(self.data['arrow'], self.data['arrow_head'])
        )

        if result is not None and result[0] == 'node':
            self.activate_node(
                result[1], result[2], exclude=(self.data['edge'].node1,)
            )

    def drop_arrow_head(self, event):
        '''The user has dropped the arrow somewhere!
        If it is on another node, make the connection.
        If it is in empty space or on the original node,
        just cancel the operation.
        '''

        self.canvas.bind('<B1-Motion>', '')
        self.canvas.bind('<ButtonRelease-1>', '')

        # Check for being near another nodes anchor point
        cx = int(self.canvas.canvasx(event.x))
        cy = int(self.canvas.canvasy(event.y))
        result = self.find_items(cx, cy)

        edge = self.data['edge']

        if result is None:
            # dropped on empty space
            self.canvas.delete('type=arrow_base')
            self.canvas.delete('type=arrow_head')
            edge.draw()
        elif result[0] == 'node':
            # dropped on another node
            node2, anchor2 = result[1:]
            if edge.node2 == node2:
                edge.anchor2 = anchor2
                edge.draw()
            else:
                # remove current connection and create new, in
                # that order -- otherwise tend to remove edge
                # completely if it is moved on same node.

                node1 = edge.node1
                anchor1 = edge.anchor1

                self.remove_edge(self.data['arrow'])

                self.add_edge(
                    node1,
                    node2,
                    'execution',
                    anchor1=anchor1,
                    anchor2=anchor2
                )

        self.data = None
        self.mouse_op = None

    def right_click_on_arrow(self, event, item, tags):
        '''Handle a right click on an arrow
        '''

        if self.popup_menu is not None:
            self.popup_menu.destroy()

        self.popup_menu = tk.Menu(self.canvas, tearoff=0)
        self.popup_menu.add_command(
            label="Delete", command=lambda: self.remove_edge(item)
        )

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def remove_edge(self, item):
        '''Remove an edge from the graph and visually
        '''

        tags = self.get_tags(item)
        edge = tags['edge']
        tag = edge.tag()
        self.graph.remove_edge(
            edge.node1, edge.node2, edge.edge_type, edge.edge_subtype
        )

        # Delete the tag, not item, so that we get all labels, etc.
        self.canvas.delete(tag)
        self.canvas.delete('type=arrow_base')
        self.canvas.delete('type=arrow_head')

    def print_edges(self, event=None):
        '''Print all the edges. Useful for debugging!
        '''

        print('All edges in tk_flowchart')
        for edge in self.edges():
            print(
                '   {} {} {} {} {} {}'.format(
                    edge.node1.tag, edge.anchor1, edge.node2.tag, edge.anchor2,
                    edge.edge_type, edge.edge_subtype
                )
            )

    def print_items(self):
        """Print all the items on the canvas, for debugging
        """

        print()
        for item in self.canvas.find_withtag('type=arrow'):
            print('{}: {}'.format(item, self.canvas.gettags(item)))

    def run(self, event=None):
        """Run the current flowchart"""

        self.update_flowchart()
        flowchart = self.flowchart.to_text()

        self._job_handler.submit_with_dialog(flowchart=flowchart)

    def push(self):
        """Save a copy of the current flowchart on the stack.
        """
        self.update_flowchart()
        self._stack.append(self.flowchart.to_dict())

    def pop(self):
        """Replace the current flowchart with the version on the stack.
        """
        self.flowchart.from_dict(self._stack.pop())
        self.from_flowchart()

    def pop_and_discard(self):
        """Remove the saved copy from the stack
        """
        self._stack.pop()

    def update_flowchart(self):
        """Update the non-graphical flowchart"""
        wf = self.flowchart

        # Make sure there is nothing in the flowchart
        wf.clear(all=True)

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in self:
            translate[node] = wf.add_node(copy.copy(node.node))
            node.update_flowchart()

        # And the edges
        for edge in self.edges():
            attr = {}
            for key in edge:
                if key not in (
                    'node1', 'node2', 'edge_type', 'edge_subtype', 'canvas'
                ):
                    attr[key] = edge[key]
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            wf.add_edge(
                node1, node2, edge.edge_type, edge.edge_subtype, **attr
            )

    def from_flowchart(self):
        """Recreate the graphics from the non-graphical flowchart"""
        wf = self.flowchart

        self.clear()

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in wf:
            extension = node.extension
            if extension is None:
                # Start node
                translate[node] = self.get_node('1')
            else:
                new_node = copy.copy(node)
                logger.debug('creating {} node'.format(extension))
                plugin = self.plugin_manager.get(extension)
                logger.debug('  plugin object: {}'.format(plugin))
                tk_node = plugin.create_tk_node(
                    tk_flowchart=self, canvas=self.canvas, node=new_node
                )
                translate[node] = tk_node
                tk_node.from_flowchart()
                self.graph.add_node(tk_node)
                tk_node.draw()

        # And the edges
        for edge in wf.edges():
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            attr = {}
            for key in edge:
                if key not in ('node1', 'node2'):
                    attr[key] = edge[key]
            self.add_edge(node1, node2, **attr)

    def _bound_to_mousewheel(self, event):
        """Set the bindings on the canvas, used when the
        mouse enters the canvas
        """
        self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)
        # self.canvas.bind_all("<Button-4>", self._on_mousewheel)
        # self.canvas.bind_all("<Button-5>", self._on_mousewheel)

    def _unbound_to_mousewheel(self, event):
        """Remove the bindings on the canvas, used when the
        mouse leaves the canvas
        """
        self.canvas.unbind_all("<MouseWheel>")
        # self.canvas.unbind_all("<Button-4>")
        # self.canvas.unbind_all("<Button-5>")

    def _on_mousewheel(self, event):
        """Handle the mousewheel or similar events.
        There are two choices for how to scroll, and it
        may differ from OS to OS.

        As set up here on a Mac the mouse drags the canvas in
        the direction of travel, thus to go down in the canvas
        you drag upwards, and vice versa.

        Flip the signs to change this
        """

        if event.num == 5 or event.delta < 0:
            delta = 1
        else:
            delta = -1

        self.canvas.yview_scroll(delta, "units")

        x0, y0, x1, y1 = self.canvas.cget('scrollregion').split(' ')
        f0, f1 = self.canvas.yview()
        y = (int(y1) - int(y0)) * f0

        tk._default_root.tk.call(self.canvas, 'moveto', self.background, 0, y)

    def xview(self, command, amount, *args):
        """Scroll in the x direction, keeping the background picture stationary
        """
        self.canvas.xview(command, amount, *args)

        x0, y0, x1, y1 = self.canvas.cget('scrollregion').split(' ')
        f0, f1 = self.canvas.xview()
        x = (int(x1) - int(x0)) * f0
        f0, f1 = self.canvas.yview()
        y = (int(y1) - int(y0)) * f0

        tk._default_root.tk.call(self.canvas, 'moveto', self.background, x, y)

    def yview(self, command, amount, *args):
        """Scroll in the y direction, keeping the background picture stationary
        """
        self.canvas.yview(command, amount, *args)

        x0, y0, x1, y1 = self.canvas.cget('scrollregion').split(' ')
        f0, f1 = self.canvas.xview()
        x = (int(x1) - int(x0)) * f0
        f0, f1 = self.canvas.yview()
        y = (int(y1) - int(y0)) * f0

        tk._default_root.tk.call(self.canvas, 'moveto', self.background, x, y)

    def clean_layout(self, event=None):
        """Clean the visual layout of the flowchart
        """

        # clear the visited flag
        for node in self:
            node.node.visited = False

        # get the node to start the traversal
        node = self.get_node('1')

        # traverse the nodes, finding what loops they are in
        self._loops = {}  # what loops a node is in
        self._in_loop = {  # ordered list of nodes directly in a loop
            'start': []
        }
        loops = tuple()

        self._loop_helper(loops, node)

        logger.debug('\nloops\n\n{}'.format(pprint.pformat(self._loops)))
        logger.debug('\nin loops\n\n{}'.format(pprint.pformat(self._in_loop)))

        # Move the nodes to the correct place on the grid
        x = 0
        y = 0
        self._loopxy = {'start': (0, 0)}
        self._layout_nodes('start', x, y)

        logger.debug('\nmax x,y\n\n{}'.format(pprint.pformat(self._loopxy)))

        # Fix the edges
        for edge in self.edges():
            # only work on edges that go upwards
            x0, y0 = edge.node1.anchor_point(edge.anchor1)
            x1, y1 = edge.node2.anchor_point(edge.anchor2)
            logger.debug(
                '   edge {}: {}, {} to {}, {}'.format(edge, x0, y0, x1, y1)
            )
            if y1 < y0:
                edge.coords = [x0, y0]
                logger.debug('   edge.node1 = {}'.format(edge.node1))
                loop = self._loops[edge.node1][-1]
                xmax, ymax = self._loopxy[loop]
                xmax = (xmax + 1) * self.grid_x
                ymax = (ymax + 1) * self.grid_y

                dx = 10 * len(self._loops[edge.node1])
                dy = 0

                # Down far enough
                edge.coords.append(x0)
                edge.coords.append(ymax - dy)
                # Right far enough
                edge.coords.append(xmax - dx)
                edge.coords.append(ymax - dy)
                # up
                edge.coords.append(xmax - dx)
                edge.coords.append(y1)
                # and to the node
                edge.coords.append(x1)
                edge.coords.append(y1)
            else:
                edge.coords = [x0, y0, x1, y1]

        # Redraw everything
        self.draw()

        del self._loops
        del self._in_loop
        del self._loopxy

    def _layout_nodes(self, loop, x, y):
        """Recursively position nodes
        """
        for node in self._in_loop[loop]:
            x0 = int(node.x)
            y0 = int(node.y)
            node.x = int((x + 0.5) * self.grid_x)
            node.y = int((y + 0.5) * self.grid_y)

            logger.debug(
                'node {} {} = {:3d} {:3d} ({:3d} {:3d}) {}'.format(
                    x, y, int(node.x), int(node.y), x0, y0, node
                )
            )

            xmax, ymax = self._loopxy[loop]
            if x > xmax:
                xmax = x
            if y > ymax:
                ymax = y
            self._loopxy[loop] = (xmax, ymax)

            if node in self._in_loop:
                self._loopxy[node] = (x + 1, y)
                x1, y = self._layout_nodes(node, x + 1, y)
                self._loopxy[loop] = (x1, y)
            else:
                y += 1

        return x, y

    def _loop_helper(self, loops, node):
        """A helper to traverse graph finding the grid locations of the nodes
        """
        node.node.visited = True
        self._loops[node] = loops

        logger.debug('node = {}, loops = {}'.format(node, loops))

        if len(loops) == 0:
            self._in_loop['start'].append(node)
        else:
            self._in_loop[loops[-1]].append(node)

        if node.node_type == 'loop':
            loops = loops + (node,)
            self._in_loop[node] = []
            # nodes in the loop
            for edge in self.graph.edges(node, direction='out'):
                if edge.edge_type == "execution" and \
                   edge.edge_subtype == 'loop':
                    node2 = edge.node2
                    if not node2.node.visited:
                        self._loop_helper(loops, node2)
            loops = loops[0:-1]
            # end exiting the loop
            for edge in self.graph.edges(node, direction='out'):
                if edge.edge_type == "execution" and \
                   edge.edge_subtype == 'exit':
                    node2 = edge.node2
                    if not node2.node.visited:
                        self._loop_helper(loops, node2)
        else:
            for edge in self.graph.edges(node, direction='out'):
                if edge.edge_type == "execution":
                    node2 = edge.node2
                    if not node2.node.visited:
                        self._loop_helper(loops, node2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The graphical interface for submitting SEAMM jobs.

A job in SEAMM is composed of a flowchart and any other files that the
flowchart requires. This module provides the TkJobHandler class, which
provides a use interface and the machinery to gather the necessary
files and submit the job to a dashboard.
"""

import configparser
import logging
from pathlib import Path
import requests
import tkinter as tk
from tkinter import messagebox
import tkinter.ttk as ttk

import Pmw

import seamm_widgets as sw

logger = logging.getLogger(__name__)


class TkJobHandler(object):

    def __init__(self, root=None):
        """Setup the Job Handler object.

        Parameters
        ----------
        root : Tk window
            The root Tk window.

        """
        self._root = root
        self.config = configparser.ConfigParser()
        self.configfile = Path.home() / '.seamm' / 'dashboards.ini'
        self.dialog = None
        self._widgets = {}

        # Read in the dashboard information if present
        self.get_configuration()
        self.current_dashboard = self.config.get(
            'global options', 'current_dashboard', fallback=None
        )
        self.timeout = self.config.get(
            'global options', 'timeout', fallback=0.5
        )

        s = ttk.Style()
        s.configure("Border.TLabel", relief='ridge', anchor=tk.W, padding=5)

    @property
    def dashboards(self):
        """The list of dashboards."""
        result = []
        for dashboard in self.config:
            if (
                dashboard
                not in ('global options', self.config.default_section)
            ):
                result.append(dashboard)
        return sorted(result)

    def submit_with_dialog(self, flowchart=None):
        """
        Allow the user to choose the dashboard and other parameters,
        and submit the job as requested.

        Parameters
        ----------
        flowchart : text, optional
            The flowchart to use. If not given, prompt the user for
            one.

        Returns
        -------
        job_id : integer
            The id of the submitted job.
        """
        if self.dialog is None:
            self.create_submit_dialog()

        result = self.dialog.activate(geometry='centerscreenfirst')

        if result is not None:
            job_id = self.submit(flowchart, **result)
            return job_id
        else:
            return None

    def submit(self, flowchart, dashboard, **job):
        """Submit the job to the given dashboard.
        """
        job['flowchart'] = flowchart
        job['path'] = 'unset'

        url = self.config[dashboard]['url']

        response = requests.post(url + '/api/jobs', json=job)
        if response.status_code != 201:
            raise RuntimeError(
                'POST /api/jobs/ {}'.format(response.status_code)
            )
        job_id = response.json()['id']
        logger.info('Submitted job #{}'.format(job_id))
        return job_id

    def create_submit_dialog(self):
        """Create the dialog for submitting a job.

        """
        logger.debug('Creating submit dialog')
        self.dialog = Pmw.Dialog(
            self._root,
            buttons=('OK', 'Cancel'),
            master=self._root,
            title='Submit job to SEAMM',
            command=self.handle_dialog
        )
        self.dialog.withdraw()

        w = self._widgets
        d = self.dialog.interior()

        # Dashboard
        dashboards = sorted(self.config.sections())
        w['dashboard'] = sw.LabeledCombobox(
            d, labeltext='Dashboard:', values=dashboards
        )
        w['dashboard'].combobox.bind("<<ComboboxSelected>>", self.dashboard_cb)

        w['add'] = ttk.Button(
            d, text='add dashboard...', command=self.add_dashboard_cb
        )

        # User and project
        w['username'] = sw.LabeledEntry(d, labeltext='User:')
        w['username'].set(self.configfile.owner())
        w['project'] = sw.LabeledCombobox(d, labeltext='Project:')

        # Title
        w['title'] = sw.LabeledEntry(d, labeltext='Title:', width=100)

        # Description
        frame = sw.ScrolledFrame(
            d, scroll_vertically=True, borderwidth=2, relief=tk.SUNKEN
        )
        f = frame.interior()
        w['description'] = tk.Text(f)
        w['description'].grid(sticky=tk.EW)

        f.rowconfigure(0, weight=1)
        f.columnconfigure(0, weight=1)

        # Set up the dashboard and projects if needed
        if len(dashboards) > 0:
            tmp = self.config.get(
                'global options', 'current_dashboard', fallback=dashboards[0]
            )
            if tmp not in dashboards:
                tmp = dashboards[0]

            if self.current_dashboard != tmp:
                self.current_dashboard = tmp
                self.save_configuration()

            w['dashboard'].set(tmp)
            self.dashboard_cb()

        # Grid the widgets into rows and columns
        w['dashboard'].grid(row=0, column=0, sticky=tk.EW)
        w['add'].grid(row=0, column=1, sticky=tk.W)
        w['username'].grid(row=1, column=0, sticky=tk.EW)
        w['project'].grid(row=1, column=1, sticky=tk.EW)
        w['title'].grid(row=2, column=0, columnspan=2, sticky=tk.W)
        frame.grid(row=3, column=0, columnspan=2, sticky=tk.NSEW)

        sw.align_labels([w['dashboard'], w['username'], w['title']])

        d.rowconfigure(3, weight=1)
        d.columnconfigure(1, weight=1)

    def handle_dialog(self, result):
        """Handle the submit dialog being completed.
        """
        if result is None or result == 'Cancel':
            self.dialog.deactivate(None)
        else:
            w = self._widgets
            self.dialog.deactivate(
                {
                    'username': w['username'].get(),
                    'project': w['project'].get(),
                    'title': w['title'].get(),
                    'dashboard': w['dashboard'].get(),
                    'description': w['description'].get(1.0, tk.END)
                }
            )

    def get_configuration(self):
        """Get the list of dashboards from the config file.
        """
        # The path to the configfile
        if self.configfile.exists():
            self.config.read(self.configfile)
        else:
            self.config.clear()

    def save_configuration(self):
        """Save the list of dashboards to disk.
        """
        # Make sure the directory exists
        self.configfile.parent.mkdir(exist_ok=True)

        # Update the current dashboard
        if self.current_dashboard is not None:
            if 'global options' not in self.config:
                self.config['global options'] = {}
            defaults = self.config['global options']
            defaults['current_dashboard'] = self.current_dashboard

        with self.configfile.open('w') as fd:
            self.config.write(fd)

    def add_dashboard_cb(self):
        """Post a dialog for adding a dashboard to the list.
        """
        dialog = Pmw.Dialog(
            self._root,
            buttons=('OK', 'Cancel'),
            master=self._root,
            title='Add Dashboard to list',
            command=self.handle_add_dialog
        )
        dialog.withdraw()
        w = self._widgets['add'] = {'dialog': dialog}

        d = dialog.interior()
        name = sw.LabeledEntry(d, labeltext='Name', width=50)
        url = sw.LabeledEntry(d, labeltext='URL')
        protocol = sw.LabeledCombobox(
            d, labeltext='Protocol', values=['http', 'sshtunnel']
        )
        protocol.set('http')

        w['name'] = name
        w['url'] = url
        w['protocol'] = protocol

        name.grid(row=0, column=0, sticky=tk.EW)
        url.grid(row=1, column=0, sticky=tk.EW)
        protocol.grid(row=2, column=0, sticky=tk.W)

        sw.align_labels([name, url, protocol])

        dialog.activate(geometry='centerscreenfirst')

    def handle_add_dialog(self, result):
        """Handle the dialog to add a dashboard to the list.
        """
        w = self._widgets['add']
        dialog = w['dialog']
        if result is None or result == 'Cancel':
            dialog.deactivate(result)
        else:
            name = w['name'].get()
            url = w['url'].get()
            protocol = w['protocol'].get()

            if name in self.config:
                messagebox.showwarning(
                    'Duplicate name', (
                        "There is already a dashboard called '{}'\n"
                        'Use a different name.'
                    ).format(name)
                )
                return

            dialog.deactivate(result)
        dialog.destroy()
        del self._widgets['add']

        # Now add to the configuration
        self.config[name] = {'url': url, 'protocol': protocol}

        self.save_configuration()

        # And reset the list in the dashboard combobox
        c = self._widgets['dashboard']
        c.combobox.config({'value': self.config.sections()})
        c.set(name)

    def dashboard_cb(self, event=None):
        """The selected dashboard has been changed"""
        w = self._widgets
        dashboard = w['dashboard'].get()

        url = self.config[dashboard]['url']

        try:
            response = requests.get(
                url + '/api/list-projects', timeout=self.timeout
            )
            if response.status_code != 200:
                logger.warning(
                    (
                        'Encountered an error getting the projects from '
                        "dashboard '{}', error code: {}"
                    ).format(dashboard, response.status_code)
                )
                messagebox.showerror(
                    title='Dashboard error',
                    message="Dashboard '{}' returned an error: {}".format(
                        dashboard, response.status_code
                    )
                )
                if self.current_dashboard is not None:
                    w['dashboard'].set(self.current_dashboard)
                return
            else:
                projects = response.json()['projects']
        except requests.exceptions.Timeout:
            logger.warning(
                'A timeout occurred contacting the dashboard ' + dashboard
            )
            messagebox.showerror(
                title='Dashboard error',
                message="Could not reach dashboard '{}'".format(dashboard)
            )
            if self.current_dashboard is not None:
                w['dashboard'].set(self.current_dashboard)
            return
        except requests.exceptions.ConnectionError:
            logger.warning(
                'A connection error occurred contacting the dashboard ' +
                dashboard
            )
            messagebox.showerror(
                title='Dashboard error',
                message=(
                    "A connection error occured reaching dashboard '{}'"
                ).format(dashboard)
            )  # yapf: disable
            if self.current_dashboard is not None:
                w['dashboard'].set(self.current_dashboard)
            return

        # All OK, changed the widgets
        w['project'].combobox.config({'value': projects})
        if len(projects) > 0:
            w['project'].set(projects[0])
        else:
            w['project'].set('')
        self.current_dashboard = dashboard
        self.save_configuration()

    def display_dashboards(self):
        """Display a list of all the dashboards with their status.

        Allow users to edit, remove and add dashboards.
        """
        # statuses = self.get_all_status(master=self._root)

        dialog = Pmw.Dialog(
            self._root,
            buttons=('OK', 'Edit', 'Remove', 'Cancel'),
            master=self._root,
            title='Dashboards',
            command=self.handle_dashboard_dialog
        )
        dialog.withdraw()
        w = self._widgets['display'] = {'dialog': dialog}

        d = dialog.interior()
        w['table'] = sw.ScrolledColumns(
            d,
            columns=[
                'Select',
                'Dashboard',
                'URL',
                'Status',
            ],
            header_style='Border.TLabel'
        )
        w['table'].grid(row=0, column=0, sticky=tk.NSEW)
        d.columnconfigure(0, weight=1)
        d.rowconfigure(0, weight=1)

        # Button to update all status
        w['update all'] = ttk.Button(
            d,
            text="Update Status of All Dashboards",
            command=self.fill_statuses
        )
        w['update all'].grid()

        # Fill in the dashboards
        table = w['table']
        f = table.interior()
        row = 0
        w['selected'] = tk.StringVar()
        if self.current_dashboard is not None:
            w['selected'].set(self.current_dashboard)
        for dashboard in self.dashboards:
            table[row, 0] = ttk.Radiobutton(
                f,
                variable=w['selected'],
                value=dashboard,
            )
            table[row, 1] = ttk.Label(f, text=dashboard, style='Border.TLabel')
            table[row, 2] = ttk.Label(
                f, text=self.config[dashboard]['url'], style='Border.TLabel'
            )
            state = self.config.get(dashboard, 'state', fallback='active')
            if state == 'active':
                table[row, 3] = ttk.Label(f, width=16, style='Border.TLabel')
            else:
                table[row, 3] = ttk.Label(
                    f, text=state, width=16, style='Border.TLabel'
                )
            table[row, 0].grid(sticky=tk.EW)
            table[row, 1].grid(sticky=tk.EW)
            table[row, 2].grid(sticky=tk.EW)
            table[row, 3].grid(sticky=tk.EW)
            row += 1

        self.fit_dialog(dialog)

        dialog.activate(geometry='centerscreenfirst')

        dialog.destroy()
        del self._widgets['display']

    def fill_statuses(self):
        w = self._widgets['display']
        dialog = w['dialog']
        table = w['table']

        dialog.configure(title='Dashboards -- Updating Status')
        progress = ttk.Progressbar(
            dialog.interior(),
            orient=tk.HORIZONTAL,
            maximum=table.nrows + 1,
            mode='determinate',
            value=1
        )
        progress.grid(sticky=tk.EW)

        for row in range(table.nrows):
            current_status = table[row, 3].cget('text')
            if current_status != 'inactive':
                table[row, 3].configure(text='...')
                table.update()
                status = self.status(table[row, 1].cget('text'))
                table[row, 3].configure(text=status)
            progress.step()
            table.update()
        progress.destroy()
        dialog.configure(title='Dashboards')

    def status(self, dashboard, timeout=1):
        """The status of the given dashboard.

        Contact the given dashboard, checking for errors, timeouts,
        etc. and report back the current status.

        Return
        ------
        status : enum (a string...)
            'up'
            'down'
            'error'
        """

        url = self.config[dashboard]['url']

        try:
            response = requests.get(url + '/api/status', timeout=timeout)
            if response.status_code != 200:
                logger.info(
                    (
                        'Encountered an error getting the status from '
                        "dashboard '{}', error code: {}"
                    ).format(dashboard, response.status_code)
                )
                result = 'dashboard error'
            else:
                result = response.text
        except requests.exceptions.Timeout:
            logger.info(
                'A timeout occurred contacting the dashboard ' + dashboard
            )
            result = 'down'
        except requests.exceptions.ConnectionError as e:
            logger.info(
                'A connection error occurred contacting the dashboard ' +
                dashboard
            )
            if e.response is not None:
                logger.info(
                    'Status code = {}, {}'.format(
                        e.response.status_code, e.response.text
                    )
                )
            result = 'connection error'

        return result

    def handle_dashboard_dialog(self, result):
        """Handle the dialog to add a dashboard to the list.
        """
        w = self._widgets['display']
        dialog = w['dialog']
        dashboard = w['selected'].get()
        if result is None or result == 'Cancel':
            dialog.deactivate(None)
        elif result == 'Edit':
            self.edit_cb(dashboard)
        elif result == 'Remove':
            self.remove_cb(dashboard)
        else:
            dialog.deactivate(None)

    def edit_cb(self, dashboard):
        """Edit the information for a dashboard.
        """
        table = self._widgets['display']['table']
        for trow in range(table.nrows):
            if table[trow, 1].cget('text') == dashboard:
                break

        dialog = Pmw.Dialog(
            self._root,
            buttons=('OK', 'Cancel'),
            title='Edit ' + dashboard.title(),
        )
        dialog.withdraw()
        w = self._widgets['edit dashboard'] = {
            'dialog': dialog,
            'dashboard': dashboard
        }

        d = dialog.interior()
        name = sw.LabeledEntry(d, labeltext='Name:', width=50)
        name.set(table[trow, 1].cget('text'))
        url = sw.LabeledEntry(d, labeltext='URL:', width=50)
        url.set(table[trow, 2].cget('text'))
        state = sw.LabeledCombobox(
            d, labeltext='State:', values=['active', 'inactive']
        )
        current_status = table[trow, 3].cget('text')
        if current_status == 'inactive':
            state.set('inactive')
        else:
            state.set('active')
        status = sw.LabeledEntry(d, labeltext='Status:', width=16)
        status.set(current_status)
        check_status = ttk.Button(
            d, text='Check Status', command=self.check_status_cb
        )
        w['status'] = status

        name.grid(row=0, columnspan=2, sticky=tk.EW)
        url.grid(row=1, columnspan=2, sticky=tk.EW)
        state.grid(row=2, columnspan=2, sticky=tk.EW)
        status.grid(row=3, sticky=tk.EW)
        check_status.grid(row=3, column=1)

        widgets = [name, url, state, status]

        # If there are any other items in the config file, put them in
        row = 3
        for key, value in self.config.items(dashboard):
            if key not in ('url', 'state', 'status'):
                row += 1
                w[key] = sw.LabeledEntry(d, labeltext=key + ':')
                w[key].set(value)
                w[key].grid(row=row, columnspan=2, sticky=tk.EW)
                widgets.append(w[key])

        sw.align_labels(widgets)

        result = dialog.activate(geometry='centerscreenfirst')

        if result == 'OK':
            if name.get() != dashboard:
                # Changed the name. Move the section in the config file
                tmp = {}
                for key, value in self.config.items(dashboard):
                    tmp[key] = value
                self.config.remove_section(dashboard)
                dashboard = name.get()
                self.config[dashboard] = tmp
                table[trow, 1].configure(text=dashboard)
                table[trow, 0].configure(value=dashboard)
                self._widgets['display']['selected'].set(dashboard)
                self.current_dashboard = dashboard

            db_config = self.config[dashboard]

            table[trow, 2].configure(text=url.get())
            db_config['url'] = url.get()
            if state.get() == 'active':
                if status.get() == 'inactive':
                    table[trow, 3].configure(text='')
                else:
                    table[trow, 3].configure(text=status.get())
            else:
                table[trow, 3].configure(text=state.get())
            db_config['state'] = state.get()

            for key, value in self.config.items(dashboard):
                if key not in ('url', 'state', 'status'):
                    db_config[key] = w[key].get()

            self.save_configuration()

    def check_status_cb(self):
        """Helper for checking the status of a dashboard."""
        w = self._widgets['edit dashboard']
        dashboard = w['dashboard']
        status = self.status(dashboard)
        w['status'].set(status)

    def fit_dialog(self, dialog):
        """Resize and fit the dialog to the current contents and the
        constraint of the window.
        """
        logger.debug('Entering fit_dialog')

        widget = self._widgets['display']['table'].interior()
        logger.debug('  widget = {}'.format(widget))
        widget.update_idletasks()

        frame = dialog.interior()
        frame.update_idletasks()
        width = frame.winfo_width()
        height = frame.winfo_height()
        sw = frame.winfo_screenwidth()
        sh = frame.winfo_screenheight()

        logger.debug(
            '  frame wxh = {} x {}, screen = {} x {}'.format(
                width, height, sw, sh
            )
        )

        mw = frame.winfo_reqwidth()
        mh = frame.winfo_reqheight()
        logger.debug('  frame requested = {} x {}'.format(mw, mh))

        # Need to handle scrolledtable using its inside frame
        ww = widget.winfo_width()
        hh = widget.winfo_height()
        w = widget.winfo_reqwidth()
        h = widget.winfo_reqheight()
        logger.debug(
            '  table wxh = {} x {}, requested = {} x {}'.format(ww, hh, w, h)
        )
        h += 20  # Add a bit of space...
        if w > mw:
            mw = w
        if h > mh:
            mh = h
        if ww > width:
            width = ww
        if hh > height:
            height = hh

        if width < mw:
            width = mw
        width += 70
        if width > 0.9 * sw:
            width = int(0.9 * sw)
        if height < mh:
            height = mh
        height += 70
        if height > 0.9 * sh:
            height = int(0.9 * sh)

        dialog.geometry('{}x{}'.format(width, height))

    def get_all_status(self, show_progress=True, master=None):
        """Get the status of all the dashboards.

        Parameter
        ---------
        show_progress : Boolean, optional
            Show a dialog with progress, default is True
        """
        if show_progress:
            dialog = tk.Toplevel(master=master)
            dialog.focus_set()  # set focus on the ProgressWindow
            dialog.grab_set()  # make a modal window
            dialog.transient(master)  # show only one window in the task bar

            dialog.title(u'Getting Status of Dashboards')
            dialog.resizable(False, False)  # window is not resizable
            # dialog.close gets fired when the window is destroyed
            # dialog.protocol(u'WM_DELETE_WINDOW', dialog.close)
            dialog.geometry('400x200+200+200')
            # cancel progress when <Escape> key is pressed
            # dialog.bind(u'<Escape>', self.close)

            progress = ttk.Progressbar(
                dialog,
                orient=tk.HORIZONTAL,
                length=len(self.dashboards) + 1,
                maximum=len(self.dashboards),
                mode='determinate',
                value=1
            )
            progress.grid(ipady=20, sticky=tk.NSEW)
            label = ttk.Label(dialog, text='Dashboard')
            label.grid()
            dialog.rowconfigure(0, minsize=30)
            dialog.columnconfigure(0, weight=1)
            dialog.update_idletasks()
            dialog.after(50)

        result = []
        for dashboard in self.dashboards:
            if show_progress:
                label.configure({'text': dashboard})
                dialog.update_idletasks()
                dialog.after(50)
            status = self.status(dashboard)
            result.append((dashboard, status))
            if show_progress:
                progress.step()
                label.configure({'text': dashboard})
                dialog.update_idletasks()
                dialog.after(50)

        if show_progress:
            master.focus_set()
            dialog.destroy()

        return result


if __name__ == "__main__":
    import configargparse
    import logging

    logger = logging.getLogger(__name__)
    parser = configargparse.ArgParser(
        auto_env_var_prefix='',
        default_config_files=[
            '/etc/seamm/seamm.ini',
            '~/.seamm/seamm.ini',
        ],
        description='Submit a SEAMM job to a dashboard'
    )

    parser.add_argument(
        '--seamm-configfile',
        is_config_file=True,
        default=None,
        help='a configuration file to override others'
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose_count",
        action="count",
        default=0,
        help="increases log verbosity for each occurence."
    )

    # Parse the configuration
    args, unknown = parser.parse_known_args()

    # Set up logging level to WARNING by default, going more verbose
    # for each new -v, to INFO and then DEBUG and finally ALL with 3 -v's

    numeric_level = max(3 - args.verbose_count, 0) * 10
    logging.basicConfig(level=numeric_level)

    # Initialize Tk
    root = tk.Tk()
    Pmw.initialise(root)

    # s = ttk.Style()
    # # print("Themes: " + str(s.theme_names()))
    # rb = ttk.Progressbar()
    # pprint.pprint(rb.configure())
    # print('Label style: ' + rb['style'])
    # class_ = rb.winfo_class()
    # print('Label class: ' + class_)
    # pprint.pprint(s.layout(class_))
    # # pprint.pprint(s.element_options('Label'))
    # # pprint.pprint(s.element_options('Label.border'))
    # # s.configure("Border.TLabel", relief='ridge', anchor=tk.W, padding=5)

    # Create the object
    job_handler = TkJobHandler(root=root)

    do_it = ttk.Button(
        root, text='Submit...', command=job_handler.submit_with_dialog
    )
    do_it.grid(sticky=tk.EW)

    status = ttk.Button(
        root, text='Dashboards...', command=job_handler.display_dashboards
    )
    status.grid(row=1, sticky=tk.EW)

    # Enter the event loop
    root.mainloop()



# -*- coding: utf-8 -*-

"""The start node in a flowchart"""

import logging
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __  # noqa: F401

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter('start')


class StartNode(seamm.Node):

    def __init__(self, flowchart=None):
        '''Initialize a specialized start node, which is the
        anchor for the graph.

        Keyword arguments:
        '''
        logger.debug('Constructing start node {}'.format(self))
        super().__init__(flowchart=flowchart, title='Start')

        self._uuid = 1

    @property
    def version(self):
        """The semantic version of this module.
        """
        return seamm.__version__

    @property
    def git_revision(self):
        """The git version of this module.
        """
        return seamm.__git_revision__

    def set_uuid(self):
        pass

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """
        return self.header + '\n'

    def describe(self):
        """Write out information about what this node will do
        """

        self.visited = True

        # The description
        job.job(self.indent + self.description_text())

        next_node = self.next()

        if next_node is None or next_node.visited:
            return None
        else:
            return next_node

    def run(self):
        """'Run' the start node, i.e. do nothing but print
        """

        next_node = super().run(printer)
        printer.important(self.header + '\n')
        return next_node

    def setup_printing(self, aprinter):
        """Establish the handlers for printing as controlled by
        options. The start step never writes to disk, so don't
        create that handler.
        """

        # First remove an existing handlers
        self.close_printing(printer)

        # A handler for stdout
        console_handler = logging.StreamHandler()
        console_handler.setLevel(printing.JOB)
        console_handler.setFormatter(self.formatter)
        printer.addHandler(console_handler)



# -*- coding: utf-8 -*-

"""The ExecLocal object does what it name implies: it executes, or
runs, an executable locally."""

import bitmath
import glob
import logging
import os
import os.path
import platform
import pprint
import pyuca
import shutil
import stat
import subprocess
import sys
import tempfile
import time

if platform.system() != 'Windows':
    import grp
    import pwd

logger = logging.getLogger(__name__)


class ExecLocal(object):

    def __init__(self):
        """Execute a flowchart, providing support for the actual
        execution of codes """

        # times for formating 'ls' like output
        self.now = int(time.time())
        self.recent = self.now - (6 * 30 * 24 * 60 * 60)  # 6 months ago

    def run(
        self,
        cmd=[],
        input_data=None,
        files=None,
        env={},
        return_files=[],
        shell=False,
    ):
        """Execute 'cmd' in a temporary directory. 'files' is a dict
        keyed by filename of files to write before execution."""

        # Create temporary directory and write the files, being
        # careful about both errors and security.

        tmpdir = tempfile.mkdtemp()
        # Ensure the file is read/write by the creator only
        saved_umask = os.umask(0o077)
        logging.info('Runnning locally in {}'.format(tmpdir))

        if files is not None:
            for filename in files:
                path = os.path.join(tmpdir, filename)
                try:
                    with open(path, "w") as fd:
                        fd.write(files[filename])
                except IOError:
                    logging.exception(
                        "An I/O error occured writing file '{}'".format(path)
                    )
                    os.umask(saved_umask)
                    shutil.rmtree(tmpdir)
                    return None
                except Exception:
                    logging.exception(
                        "An unexpected error occured writing file '{}'"
                        .format(path)
                    )
                    os.remove(path)
                    os.umask(saved_umask)
                    shutil.rmtree(tmpdir)
                    return None

        os.umask(saved_umask)

        # Now execute the program in the temp directory
        logger.debug('about to run ' + ' '.join(cmd))

        p = subprocess.run(
            cmd,
            input=input_data,
            env=dict(os.environ, **env),
            cwd=tmpdir,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell
        )

        logger.debug('\n' + pprint.pformat(p))

        # capture the return code and output
        result = {
            'returncode': p.returncode,
            'stdout': p.stdout,
            'stderr': p.stderr,
        }

        # capture the list of files in the directory
        c = pyuca.Collator()
        listing = ''
        self.now = int(time.time())
        self.recent = self.now - (6 * 30 * 24 * 60 * 60)  # 6 months ago
        for dirpath, dirs, files in os.walk(tmpdir):
            # Do locale sensitive sort of files to list
            listing += dirpath + '\n' + '\n\t'.join(
                self.ls_format(dirpath, sorted(files, key=c.sort_key))
            )
        result['listing'] = listing

        # capture the requested files
        result['files'] = []
        for pattern in return_files:
            filenames = glob.glob(os.path.join(tmpdir, pattern))
            for path in filenames:
                filename = os.path.basename(path)
                data = None
                exception = None
                result['files'].append(filename)
                try:
                    with open(path, "r") as fd:
                        data = fd.read()
                except IOError:
                    exception = sys.exc_info()
                    logging.warning(
                        "An I/O error occured reading file '{}'"
                        .format(filename),
                        exc_info=exception
                    )
                except Exception:
                    exception = sys.exc_info()
                    logging.warning(
                        "An unexpected error occured reading file '{}'"
                        .format(filename),
                        exc_info=exception
                    )
                finally:
                    result[filename] = {'exception': exception, 'data': data}

        # Clean up the temporary directory
        shutil.rmtree(tmpdir)

        return result

    def get_mode_info(self, filename, mode):
        """Get the mode information for 'ls' like listing"""

        perms = "-"
        link = ""

        if stat.S_ISDIR(mode):
            perms = "d"
        elif stat.S_ISLNK(mode):
            perms = "l"
            link = os.readlink(filename)
        mode = stat.S_IMODE(mode)
        for who in "USR", "GRP", "OTH":
            for what in "R", "W", "X":
                # lookup attributes at runtime using getattr
                if mode & getattr(stat, "S_I" + what + who):
                    perms = perms + what.lower()
                else:
                    perms = perms + "-"
        # return multiple bits of info in a tuple
        return (perms, link)

    def ls_format(self, path, files):
        """Format a list of files as in 'ls'"""

        result = []
        for filename in files:
            try:  # exceptions
                # Get all the file info
                stat_info = os.lstat(os.path.join(path, filename))
            except Exception:
                result.append("{}: No such file or directory".format(filename))
                continue

            perms, link = self.get_mode_info(
                os.path.join(path, filename), stat_info.st_mode
            )

            nlink = "%4d" % stat_info.st_nlink  # formatting strings

            if platform.system() == 'Windows':
                name = 8 * ' '
                group = 8 * ' '
            else:
                try:
                    name = "%-8s" % pwd.getpwuid(stat_info.st_uid)[0]
                except KeyError:
                    name = "%-8s" % stat_info.st_uid

                try:
                    group = "%-8s" % grp.getgrgid(stat_info.st_gid)[0]
                except KeyError:
                    group = "%-8s" % stat_info.st_gid

            size = bitmath.Byte(stat_info.st_size).best_prefix().format(
                "{value:.1f} {unit}"
            )  # noqa: E124

            # Get time stamp of file
            ts = stat_info.st_mtime
            if (ts < self.recent) or (ts > self.now):  # boolean operators
                time_fmt = "%b %e  %Y"
            else:
                time_fmt = "%b %e %R"
            time_str = time.strftime(time_fmt, time.gmtime(ts))

            # Format the result
            tmp = '{} {} {} {} {} {} {}'.format(
                perms, nlink, name, group, size, time_str, filename
            )

            if link:
                tmp += " -> " + link

            result.append(tmp)

        return result



# -*- coding: utf-8 -*-

"""The ExecWorkflow object does what it name implies: it executes, or
runs, a given flowchart.

It provides the environment for running the computational tasks
locally or remotely, using what is commonly called workflow management
system (WMS).  The WMS concept, as used here, means tools that run
given tasks without knowing anything about chemistry. The chemistry
specialization is contained in the Flowchart and the nodes that it
contains."""

import logging
import os.path
import sys
import traceback

import molsystem
import reference_handler
import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __  # noqa: F401

logger = logging.getLogger(__name__)
job = printing.getPrinter()


class ExecFlowchart(object):

    def __init__(self, flowchart=None):
        """Execute a flowchart, providing support for the actual
        execution of codes """

        self.flowchart = flowchart

    def run(self, root=None):
        if not self.flowchart:
            raise RuntimeError('There is no flowchart to run!')

        # Create the global context
        logger.debug('Creating global variables space')
        seamm.flowchart_variables = seamm.Variables()

        # Create the system handler and default system in the global context
        systems = molsystem.Systems()
        system = systems.create_system('seamm')
        seamm.flowchart_variables.set_variable('_systems', systems)
        seamm.flowchart_variables.set_variable('_system', system)

        self.flowchart.root_directory = root

        # Correctly number the nodes
        self.flowchart.set_ids()

        # Write out an initial summary of the flowchart before doing anything
        # Reset the visited flag for traversal
        self.flowchart.reset_visited()

        # Get the start node
        next_node = self.flowchart.get_node('1')

        # describe ourselves
        job.job(
            (
                '\nDescription of the flowchart'
                '\n----------------------------'
            )
        )

        while next_node:
            try:
                next_node = next_node.describe()
            except Exception:
                message = (
                    'Error describing the flowchart\n\n' +
                    traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise
            except:  # noqa: E722
                message = (
                    'Unexpected error describing the flowchart\n\n' +
                    traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise

        job.job('')

        # And actually run it!
        job.job(('Running the flowchart\n' '---------------------'))

        next_node = self.flowchart.get_node('1')
        while next_node is not None:
            try:
                next_node = next_node.run()
            except DeprecationWarning as e:
                print('\nDeprecation warning: ' + str(e))
                traceback.print_exc(file=sys.stderr)
                traceback.print_exc(file=sys.stdout)
            except Exception:
                message = (
                    'Error running the flowchart\n\n' + traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                break
            except:  # noqa: E722
                message = (
                    'Unexpected error running the flowchart\n\n' +
                    traceback.format_exc()
                )
                print(message)
                logger.critical(message)
                raise

        # Write the final structure
        system = seamm.flowchart_variables.get_variable('_system')
        if system.n_atoms() > 0:
            # MMCIF file has bonds
            filename = os.path.join(
                self.flowchart.root_directory, 'final_structure.mmcif'
            )
            text = None
            try:
                text = system.to_mmcif_text()
            except Exception:
                message = (
                    'Error creating the final mmcif file\n\n' +
                    traceback.format_exc()
                )
                print(message)
                logger.critical(message)

            if text is not None:
                with open(filename, 'w') as fd:
                    print(text, file=fd)
                job.job(
                    "\nWrote the final structure to 'final_structure.mmcif' "
                    'for viewing.'
                )
            # CIF file has cell
            if system.periodicity == 3:
                filename = os.path.join(
                    self.flowchart.root_directory, 'final_structure.cif'
                )
                with open(filename, 'w') as fd:
                    print(system.to_cif_text(), file=fd)
                job.job(
                    "\nWrote the final structure to 'final_structure.cif' for "
                    'viewing.'
                )

        # And print out the references
        filename = os.path.join(self.flowchart.root_directory, 'references.db')
        try:
            references = reference_handler.Reference_Handler(filename)
        except Exception as e:
            job.job('Error with references:')
            job.job(e)

        if references.total_citations() > 0:
            tmp = {}
            citations = references.dump(fmt='text')
            for citation, text, count, level in citations:
                if level not in tmp:
                    tmp[level] = {}
                tmp[level][citation] = (text, count)

            for level in sorted(tmp.keys()):
                ref_dict = tmp[level]
                if level == 1:
                    job.job('\nPrimary references:\n')
                elif level == 2:
                    job.job('\nSecondary references:\n')
                else:
                    job.job('\nLess important references:\n')

                lines = []
                for citation in sorted(ref_dict.keys()):
                    text, count = ref_dict[citation]
                    if count == 1:
                        lines.append('({}) {:s}'.format(citation, text))
                    else:
                        lines.append(
                            '({}) {:s} (used {:d} times)'.format(
                                citation, text, count
                            )
                        )
                job.job(
                    __(
                        '\n\n'.join(lines),
                        indent=4 * ' ',
                        indent_initial=False
                    )
                )
        # Close the reference handler, which should force it to close the
        # connection.
        del references



# -*- coding: utf-8 -*-

import collections.abc
import copy
import logging
import json
import Pmw
import seamm
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk
"""A graphical node using Tk on a canvas"""

logger = logging.getLogger(__name__)


class TkNode(collections.abc.MutableMapping):
    """The abstract base class for all Tk-based nodes"""

    anchor_points = {
        's': (+0.00, +0.50),
        'sse': (+0.25, +0.50),
        'se': (+0.50, +0.50),
        'ese': (+0.50, +0.25),
        'e': (+0.50, +0.00),
        'ene': (+0.50, -0.25),
        'ne': (+0.50, -0.50),
        'nne': (+0.25, -0.50),
        'n': (+0.00, -0.50),
        'nnw': (-0.25, -0.50),
        'nw': (-0.50, -0.50),
        'wnw': (-0.50, -0.25),
        'w': (-0.50, +0.00),
        'wsw': (-0.50, +0.25),
        'sw': (-0.50, +0.50),
        'ssw': (-0.25, +0.50)
    }

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        node_type='simple',
        canvas=None,
        x=None,
        y=None,
        w=None,
        h=None,
        my_logger=logger,
        keyword_metadata=None,
        keywords=None
    ):
        """Initialize a node

        Keyword arguments:
        """
        self.logger = my_logger
        self.tk_flowchart = tk_flowchart
        self.tk_subflowchart = None
        self.node = node
        self._keyword_metadata = keyword_metadata
        self.toplevel = None
        self.canvas = canvas

        if self.node is not None:
            if self.node.x is None:
                self.node.x = x
            if self.node.y is None:
                self.node.y = y
            if self.node.w is None:
                self.node.w = w
            if self.node.h is None:
                self.node.h = h

        self.node_type = node_type

        self._border = None
        self.title_label = None
        self._selected = False
        self.popup_menu = None
        self._tmp = None
        self.dialog = None
        self.previous_grab = None

        # Widget information
        self._widget = {}
        self.tk_var = {}
        self.results_widgets = None

    def __hash__(self):
        """Make iterable!"""
        return self.node.uuid

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and
            self.__hash__() == other.__hash__()
        )

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets!"""
        return self._widget[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._widget[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self._widget:
            self._widget[key].destroy()
        del self._widget[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._widget)

    def __len__(self):
        """The len() command"""
        return len(self._widget)

    @property
    def uuid(self):
        """The uuid of the node"""
        return self.node.uuid

    @property
    def title(self):
        """The title to display"""
        return self.node.title

    @title.setter
    def title(self, value):
        self.node.title = value
        if self.title_label is not None:
            self.canvas.itemconfigure(self.title_label, text=value)

    @property
    def tag(self):
        """The string representation of the uuid of the node"""
        return self.node.tag

    @property
    def flowchart(self):
        """The flowchart object"""
        return self.node.flowchart

    @flowchart.setter
    def flowchart(self, value):
        """The flowchart object"""
        self.node.flowchart = value

    @property
    def x(self):
        """The x-position of the center of the graphical node"""
        return self.node.x

    @x.setter
    def x(self, value):
        self.node.x = value

    @property
    def y(self):
        """The y-position of the center of the graphical node"""
        return self.node.y

    @y.setter
    def y(self, value):
        self.node.y = value

    @property
    def w(self):
        """The width of the graphical node"""
        return self.node.w

    @w.setter
    def w(self, value):
        self.node.w = value

    @property
    def h(self):
        """The height of the graphical node"""
        return self.node.h

    @h.setter
    def h(self, value):
        self.node.h = value

    def set_uuid(self):
        self.node.set_uuid()

    def connections(self):
        """Return a list of all the incoming and outgoing edges
        for this node, giving the anchor points and other node
        """

        return self.tk_flowchart.edges(self)

    @property
    def selected(self):
        """Whether I am selected or not"""
        return self._selected

    @selected.setter
    def selected(self, value):
        self._selected = value
        if value:
            self.canvas.itemconfigure(self.border, outline='red')
        else:
            self.canvas.itemconfigure(self.border, outline='black')

    @property
    def canvas(self):
        """The canvas for drawing the node"""
        return self._canvas

    @canvas.setter
    def canvas(self, value):
        if value:
            self.toplevel = value.winfo_toplevel()
        self._canvas = value

    @property
    def border(self):
        """The border of the picture in the flowchart"""
        return self._border

    @border.setter
    def border(self, value):
        self._border = value

    def draw(self):
        """Draw the node on the given canvas, making it visible"""

        # Remove any graphics items
        self.undraw()

        # the outline
        x0 = self.x - self.w / 2
        x1 = x0 + self.w
        y0 = self.y - self.h / 2
        y1 = y0 + self.h
        self._border = self.canvas.create_rectangle(
            x0,
            y0,
            x1,
            y1,
            tags=[self.tag, 'type=outline'],
            fill='white',
        )

        # the label in the middle
        self.title_label = self.canvas.create_text(
            self.x, self.y, text=self.title, tags=[self.tag, 'type=title']
        )

        for direction, edge in self.connections():
            edge.move()

    def undraw(self):
        """Remove all of our visual components
        """

        self.canvas.delete(self.tag)

    def move(self, deltax, deltay):
        if self._tmp is None:
            self._tmp = self.connections()

        self.x += deltax
        self.y += deltay

        self.canvas.move(self.tag, deltax, deltay)

        for connection in self._tmp:
            direction, edge = connection
            edge.move()

    def end_move(self, deltax, deltay):
        self.move(deltax, deltay)
        self._x0 = None
        self._y0 = None
        self._tmp = None

    def right_click(self, event):
        """Do whatever needs to be done for a right-click on this
        item in the flowchart.

        Subclasses should override this as appropriate! The menu
        created in this base method is accessible in subclasses
        which should make it relatively easy to override.
        """

        if self.popup_menu is not None:
            self.popup_menu.destroy()

        self.popup_menu = tk.Menu(self.canvas, tearoff=0)
        self.popup_menu.add_command(
            label="Delete",
            command=lambda: self.tk_flowchart.remove_node(self)
        )

        if type(self) is seamm.tk_node.TkNode:
            self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def double_click(self, event):
        """Do whatever needs to be done for a double-click on this
        item in the flowchart.

        Subclasses should override this as appropriate!
        """

        self.edit()

    def activate(self):
        """Add active handles at the anchor points and change the
        cursor
        """

        self.canvas.delete(self.tag + ' && type=anchor')
        for pt, x, y in self.anchor_point("all"):
            x0 = x - 2
            y0 = y - 2
            x1 = x + 2
            y1 = y + 2
            self.canvas.create_oval(
                x0,
                y0,
                x1,
                y1,
                fill='red',
                outline='red',
                tags=[self.tag, 'type=anchor', 'anchor=' + pt]
            )

    def deactivate(self):
        """Remove the decorations indicate the anchor points
        """

        self.canvas.delete(self.tag + ' && type=anchor')
        self.canvas.delete(self.tag + ' && type=active_anchor')

    def anchor_point(self, anchor="all"):
        """Where the anchor points are located. If "all" is given
        a dictionary of all points is returned"""

        if anchor == "all":
            result = []
            for pt in type(self).anchor_points:
                a, b = type(self).anchor_points[pt]
                result.append(
                    (pt, int(self.x + a * self.w), int(self.y + b * self.h))
                )
            return result

        if anchor in type(self).anchor_points:
            a, b = type(self).anchor_points[anchor]
            return (int(self.x + a * self.w), int(self.y + b * self.h))

        raise NotImplementedError(
            "anchor position '{}' not implemented".format(anchor)
        )

    def check_anchor_points(self, x, y, halo):
        """If the position x, y is within halo or one of the anchor points
        activate the point and return the name of the anchor point
        """

        points = []
        for direction, edge in self.connections():
            if direction == 'out':
                points.append(edge.anchor1)
            else:
                points.append(edge.anchor2)

        for point, x0, y0 in self.anchor_point():
            if x >= x0 - halo and x <= x0 + halo and \
               y >= y0 - halo and y <= y0 + halo:
                if point in points:
                    return None
                else:
                    return point
        return None

    def is_inside(self, x, y, halo=0):
        """Return a boolean indicating whether the point x, y is inside
        this node, using halo as a size around the point
        """
        if x < self.x - self.w / 2 - halo:
            return False
        if x > self.x + self.w / 2 + halo:
            return False

        if y < self.y - self.h / 2 - halo:
            return False
        if y > self.y + self.h / 2 + halo:
            return False

        return True

    def activate_anchor_point(self, point, halo):
        """Put a marker on the anchor point to indicate it is
        active
        """

        x, y = self.anchor_point(point)
        self.canvas.create_oval(
            x - halo,
            y - halo,
            x + halo,
            y + halo,
            fill='red',
            outline='red',
            tags=[self.tag, 'type=active_anchor', 'anchor=' + point]
        )

    def remove_edge(self, edge):
        """Remove a given edge, or all edges if 'all' is given
        """

        if isinstance(edge, str) and edge == 'all':
            for direction, obj in self.connections():
                self.remove_edge(obj)
        else:
            self.tk_flowchart.graph.remove_edge(
                edge.node1, edge.node2, edge.edge_type, edge.edge_subtype
            )

    def edit(self):
        """Present a dialog for editing this step's parameters.

        Subclasses can override this.
        """
        # Create the dialog if it doesn't exist
        if self.dialog is None:
            self.create_dialog()
            # After full creation, reset the dialog. This may do nothing,
            # or may layout the widgets, but can only be done after fully
            # creating the dialog.
            self.reset_dialog()
            # And resize the dialog to fit...
            self.fit_dialog()

        # And put it on-screen, the first time centered. If it contains
        # a subflowchart, save it so it can be restored on a 'Cancel'
        if self.tk_subflowchart is not None:
            self.tk_subflowchart.push()

        self.dialog.activate(geometry='centerscreenfirst')

    def create_dialog(
        self,
        title='Edit step',
        widget='frame',
        results_tab=False,
    ):
        """Create the base dialog for editing the parameters for a step.

        At the moment I have removed the Help button.
        """
        self.logger.debug('Create dialog in tk_node base class')
        self.dialog = Pmw.Dialog(
            self.toplevel,
            buttons=('OK', 'Cancel'),
            master=self.toplevel,
            title=title,
            command=self.handle_dialog
        )
        self.dialog.withdraw()

        if widget == 'frame':
            # Create a frame to hold everything
            frame = ttk.Frame(self.dialog.interior())
            frame.pack(expand=tk.YES, fill=tk.BOTH)
            self['frame'] = frame
            return frame
        elif (
            widget == 'notebook' or results_tab or
            self._keyword_metadata is not None
        ):
            # A tabbed notebook
            notebook = ttk.Notebook(self.dialog.interior())
            notebook.pack(side='top', fill=tk.BOTH, expand=tk.YES)
            self['notebook'] = notebook

            # Main frame holding the widgets
            frame = ttk.Frame(notebook)
            self['frame'] = frame
            notebook.add(frame, text='Parameters', sticky=tk.NW)

        if results_tab:
            # Second tab for results if requested
            rframe = self['results frame'] = ttk.Frame(notebook)
            notebook.add(rframe, text='Results', sticky=tk.NSEW)

            # Shortcut for parameters
            P = self.node.parameters

            var = self.tk_var['create tables'] = tk.IntVar()
            if P['create tables'].value == 'yes':
                var.set(1)
            else:
                var.set(0)
            self['create tables'] = ttk.Checkbutton(
                rframe, text='Create tables if needed', variable=var
            )
            self['create tables'].grid(row=0, column=0, sticky=tk.W)

            self['results'] = sw.ScrolledColumns(
                rframe,
                columns=[
                    'Result',
                    'Save',
                    'Variable name',
                    'In table',
                    'Column name',
                ]
            )
            self['results'].grid(row=1, column=0, sticky=tk.NSEW)
            rframe.columnconfigure(0, weight=1)
            rframe.rowconfigure(1, weight=1)

        if self._keyword_metadata is not None:
            # Next tab to handle adding keywords manually
            self.logger.debug('Adding the keyword tab')
            kframe = self['add_to_input'] = ttk.Frame(notebook)
            notebook.add(kframe, text='Add to input', sticky=tk.NSEW)
            self['keywords'] = sw.Keywords(
                kframe,
                metadata=self._keyword_metadata,
                keywords=self.node.parameters['extra keywords'].value
            )
            self['keywords'].pack(expand='yes', fill='both')

        return frame

    def setup_results(self, properties, calculation='energy'):
        """Layout the results tab of the dialog"""
        results = self.node.parameters['results'].value

        self.results_widgets = []
        table = self['results']
        frame = table.interior()

        row = 0
        for key, entry in properties.items():
            if 'calculation' not in entry:
                continue
            if calculation not in entry['calculation']:
                continue
            if 'dimensionality' not in entry:
                continue
            if entry['dimensionality'] != 'scalar':
                continue

            widgets = []
            widgets.append(key)

            table.cell(row, 0, entry['description'])

            # variable
            var = self.tk_var[key] = tk.IntVar()
            var.set(0)
            w = ttk.Checkbutton(frame, variable=var)
            table.cell(row, 1, w)
            widgets.append(w)
            e = ttk.Entry(frame, width=15)
            e.insert(0, key.lower())
            table.cell(row, 2, e)
            widgets.append(e)

            if key in results:
                if 'variable' in results[key]:
                    var.set(1)
                    e.delete(0, tk.END)
                    e.insert(0, results[key]['variable'])

            # table
            w = ttk.Combobox(frame, width=10)
            table.cell(row, 3, w)
            widgets.append(w)
            e = ttk.Entry(frame, width=15)
            e.insert(0, key.lower())
            table.cell(row, 4, e)
            widgets.append(e)

            if key in results:
                if 'table' in results[key]:
                    w.set(results[key]['table'])
                    e.delete(0, tk.END)
                    e.insert(0, results[key]['column'])

            self.results_widgets.append(widgets)
            row += 1

    def fit_dialog(self):
        """Resize and fit the dialog to the current contents and the
        constraint of the window.
        """
        self.logger.debug('Entering fit_dialog')
        frame = self['frame']
        frame.update_idletasks()
        width = frame.winfo_width()
        height = frame.winfo_height()
        sw = frame.winfo_screenwidth()
        sh = frame.winfo_screenheight()

        self.logger.debug(
            '  frame wxh = {} x {}, screen = {} x {}'.format(
                width, height, sw, sh
            )
        )

        mw = 0
        mh = 0
        if 'notebook' in self:
            for tab in self['notebook'].tabs():
                widget = frame.nametowidget(tab)
                widget.update_idletasks()
                self.logger.debug('  widget = {}'.format(widget))
                ww = widget.winfo_width()
                hh = widget.winfo_height()
                w = widget.winfo_reqwidth()
                h = widget.winfo_reqheight()
                self.logger.debug(
                    '  tab {} wxh = {} x {}, requested = {} x {}'.format(
                        tab, ww, hh, w, h
                    )
                )
                if w > mw:
                    mw = w
                if h > mh:
                    mh = h
                if ww > width:
                    width = ww
                if hh > height:
                    height = hh
            # Need to do results again using the inside of the scrolled table..
            if 'results' in self:
                widget = self['results'].interior()
                self.logger.debug('  widget = {}'.format(widget))
                widget.update_idletasks()
                ww = widget.winfo_width()
                hh = widget.winfo_height()
                w = widget.winfo_reqwidth()
                h = widget.winfo_reqheight()
                self.logger.debug(
                    '  tab {} wxh = {} x {}, requested = {} x {}'.format(
                        tab, ww, hh, w, h
                    )
                )
                if w > mw:
                    mw = w
                if h > mh:
                    mh = h
                if ww > width:
                    width = ww
                if hh > height:
                    height = hh
        else:
            mw = frame.winfo_reqwidth()
            mh = frame.winfo_reqheight()
            self.logger.debug('  frame requested = {} x {}'.format(mw, mh))

        if width < mw:
            width = mw
        width += 70
        if width + 70 > 0.9 * sw:
            width = int(0.9 * sw)
        if height < mh:
            height = mh
        height += 70
        if height > 0.9 * sh:
            height = int(0.9 * sh)

        self.dialog.geometry('{}x{}'.format(width, height))

    def reset_dialog(self, widget=None):
        """Reset the layout of the dialog as needed for the parameters.

        In this base class this does nothing. Override as needed in the
        subclasses derived from this class.
        """
        pass

    def handle_dialog(self, result):
        """Do the right thing when the dialog is closed.
        """
        if result is None or result == 'Cancel':
            self.dialog.deactivate(result)

            # If there is a subflowchart, revert to the saved copy
            if self.tk_subflowchart is not None:
                self.tk_subflowchart.pop()

            # Reset the results widgets if they exist
            if self.results_widgets is not None:
                results = self.node.parameters['results']['value']
                self.logger.debug('Resetting results on Cancel')
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug('  results dict\n---------')
                    for key, item in results.items():
                        self.logger.debug(key)
                        self.logger.debug(
                            json.dumps(results[key], sort_keys=True, indent=3)
                        )

                for key, w_check, w_variable, w_table, w_column in self.results_widgets:  # noqa: E501
                    self.logger.debug('  key: {}'.format(key))
                    w_variable.delete(0, tk.END)
                    w_table.delete(0, tk.END)
                    w_column.delete(0, tk.END)
                    if key in results:
                        tmp = results[key]
                        self.logger.debug(
                            '  key dict\n---------\n' +
                            json.dumps(tmp, sort_keys=True, indent=3) +
                            '\n-----'
                        )
                        if 'variable' in tmp:
                            self.tk_var[key].set(1)
                            w_variable.insert(0, tmp['variable'])
                        else:
                            self.tk_var[key].set(0)
                            w_variable.insert(0, key.lower())

                        if 'table' in tmp:
                            w_table.insert(0, tmp['table'])
                            w_column.insert(0, tmp['column'])
                        else:
                            w_table.set('')
                            w_column.insert(0, key.lower())
                    else:
                        self.logger.debug('  resetting widgets')
                        self.tk_var[key].set(0)
                        w_variable.insert(0, key.lower())
                        w_column.insert(0, key.lower())

            # Reset the parameters, if any
            if self.node.parameters is not None:
                self.node.parameters.reset_widgets()

            # Reset any keywords
            if 'keywords' in self:
                self['keywords'].reset()

            # Reset the layout to make sure it is correct
            self.reset_dialog()

        elif result == 'Help':
            self.help()
        elif result == 'OK':
            self.dialog.deactivate(result)

            # Capture the parameters from the widgets
            if self.node.parameters is not None:
                self.node.parameters.set_from_widgets()

            # If there is a subflowchart, throw the saved copy away
            if self.tk_subflowchart is not None:
                self.tk_subflowchart.pop_and_discard()

            # Get what results to store, if the results tab exists
            if self.results_widgets is not None:
                # Shortcut for parameters
                P = self.node.parameters

                # and from the results tab...
                if self.tk_var['create tables'].get():
                    P['create tables'].value = 'yes'
                else:
                    P['create tables'].value = 'no'

                results = P['results'].value = {}
                for key, w_check, w_variable, w_table, w_column in self.results_widgets:  # noqa: E501

                    if self.tk_var[key].get():
                        tmp = results[key] = dict()
                        tmp['variable'] = w_variable.get()
                    table = w_table.get()
                    if table != '':
                        if key not in results:
                            tmp = results[key] = dict()
                        tmp['table'] = table
                        tmp['column'] = w_column.get()
            # And any keywords
            if 'keywords' in self:
                P = self.node.parameters
                P['extra keywords'].value = self['keywords'].get_keywords()
                self['keywords'].keywords = P['extra keywords'].value
        else:
            self.dialog.deactivate(result)
            raise RuntimeError(
                "Don't recognize dialog result '{}'".format(result)
            )

    def help(self):
        """Base class for presenting help, does nothing.

        Subclasses should override this.
        """
        pass

    def to_dict(self):
        """Serialize to a dict"""
        data = {
            'x': self._x,
            'y': self._y,
            'w': self._w,
            'h': self._h,
        }

        return data

    def update_flowchart(self, tk_flowchart=None, flowchart=None):
        """Update the nongraphical flowchart. Only used in nodes that contain
        flowcharts"""
        if tk_flowchart is None or flowchart is None:
            return

        # Make sure there is nothing in the flowchart
        flowchart.clear(all=True)

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in tk_flowchart:
            translate[node] = flowchart.add_node(copy.copy(node.node))
            node.update_flowchart()

        # And the edges
        for edge in tk_flowchart.edges():
            attr = {}
            for key in edge:
                if key not in (
                    'node1', 'node2', 'edge_type', 'edge_subtype', 'canvas'
                ):
                    attr[key] = edge[key]
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            flowchart.add_edge(
                node1, node2, edge.edge_type, edge.edge_subtype, **attr
            )

    def from_flowchart(self, tk_flowchart=None, flowchart=None):
        """Recreate the graphics from the non-graphical flowchart.
        Only used in nodes that contain flowchart"""

        if tk_flowchart is None or flowchart is None:
            return

        tk_flowchart.clear()

        # Add all the non-graphical nodes, making copies so that
        # when the flowchart is cleared our objects still exist
        translate = {}
        for node in flowchart:
            extension = node.extension
            if extension is None:
                # Start node
                translate[node] = tk_flowchart.get_node('1')
            else:
                new_node = copy.copy(node)
                self.logger.debug('creating {} node'.format(extension))
                plugin = tk_flowchart.plugin_manager.get(extension)
                self.logger.debug('  plugin object: {}'.format(plugin))
                tk_node = plugin.create_tk_node(
                    tk_flowchart=tk_flowchart,
                    canvas=tk_flowchart.canvas,
                    node=new_node
                )
                translate[node] = tk_node
                tk_node.from_flowchart()
                tk_flowchart.graph.add_node(tk_node)
                tk_node.draw()

        # And the edges
        for edge in flowchart.edges():
            node1 = translate[edge.node1]
            node2 = translate[edge.node2]
            attr = {}
            for key in edge:
                if key not in ('node1', 'node2'):
                    attr[key] = edge[key]
            tk_flowchart.add_edge(node1, node2, **attr)

    def default_edge_subtype(self):
        """Return the default subtype of the edge. Usually this is ''
        but for nodes with two or more edges leaving them, such as a loop, this
        method will return an appropriate default for the current edge. For
        example, by default the first edge emanating from a loop-node is the
        'loop' edge; the second, the 'exit' edge.

        A return value of 'too many' indicates that the node exceeds the number
        of allowed exit edges.
        """

        # how many outgoing edges are there?
        n_edges = len(self.tk_flowchart.edges(self, direction='out'))

        self.logger.debug(
            'node.default_edge_label, n_edges = {}'.format(n_edges)
        )

        if n_edges == 0:
            return "next"
        else:
            return "too many"

    def next_anchor(self):
        """Return where the next node should be positioned. The default is
        <gap> below the 's' anchor point.
        """

        return 's'



# -*- coding: utf-8 -*-

"""Control parameters for a step in a MolSSI flowchart"""

import collections.abc
from distutils.util import strtobool
import importlib
import json
import logging
from seamm_util import Q_
from seamm_util import ureg
import pprint

logger = logging.getLogger(__name__)

# All for a default root context for evaluating expressions
# and variables

root_context = None


def set_context(context):
    """Set the default root context for evaluating variables
    and expressions in parameters."""

    global root_context
    root_context = context


class Parameter(collections.abc.MutableMapping):
    """A single parameter, with defaults, units, description, etc.
    This is object is a dict-like mutable mapping with properties
    to make it appear to be a simple object with attributes.
    """

    def __init__(self, *args, **kwargs):
        """Initialize this parameter"""

        logger.debug('\nParameter.__init__')

        self._data = {}
        self.dimensionality = None
        self._widget = None

        self.reset()

        # Handle positional or keyword arguments
        for data in args:
            if isinstance(data, dict):
                self.update(data)
            else:
                raise RuntimeError('Positional arguments must be dicts')

        self.update(kwargs)

        logger.debug('Finished constructing Parameter\n')

    def __getitem__(self, key):
        """Allow [] access to the dictionary!"""
        return self._data[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._data[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        del self._data[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._data)

    def __len__(self):
        """The len() command"""
        return len(self._data)

    def __repr__(self):
        """The official string representation of this object"""
        if self.units is None or self.units == '':
            return self.value
        else:
            return ('{} {}').format(self.value, self.units)

    def __str__(self):
        if self.units is None or self.units == '':
            if self.kind == 'integer':
                try:
                    value = int(self.value)
                    return ('{:' + self.format_string + '}').format(value)
                except Exception:
                    return ('{}').format(self.value)
            if self.kind == 'float':
                try:
                    value = float(self.value)
                    return ('{:' + self.format_string + '}').format(value)
                except ValueError:
                    return ('{}').format(self.value)
            if self.format_string == '':
                return str(self.value)
            else:
                return ('{:' + self.format_string + '}').format(self.value)
        else:
            if self.kind == 'integer':
                try:
                    value = int(self.value)
                    return ('{:' + self.format_string +
                            '} {}').format(value, self.units)
                except ValueError:
                    return ('{} {}').format(self.value, self.units)
            if self.kind == 'float':
                try:
                    value = float(self.value)
                    return ('{:' + self.format_string +
                            '} {}').format(value, self.units)
                except Exception:
                    return ('{} {}').format(self.value, self.units)
            if self.format_string == '':
                return '{} {}'.format(self.value, self.units)
            else:
                return ('{:' + self.format_string +
                        '} {}').format(self.value, self.units)

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        if item in self._data:
            return True
        return False

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self._data == other._data

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self._data.copy()

    @property
    def value(self):
        """The current value of the parameter. May be a value, a
        Python expression containing variables prefix with $,
        standard operators or parentheses."""

        if 'value' not in self._data:
            self._data['value'] = self._data['default']

        result = self._data['value']
        if result is None:
            result = self._data['default']

        return result

    @value.setter
    def value(self, value):
        self._data['value'] = value

    @property
    def default(self):
        """The current default of the parameter. May be a value, a
        Python expression containing variables prefix with $,
        standard operators or parenthesise, or a pint units
        quantity."""

        return self._data['default']

    @default.setter
    def default(self, value):
        self._data['default'] = value

    @property
    def kind(self):
        """The type of the parameter: integer, float, string,
        enum or special.
        This can be used to convert the value to the correct
        type in e.g. get_value."""

        return self._data['kind']

    @kind.setter
    def kind(self, value):
        if value not in ('integer', 'float', 'string'):
            raise RuntimeError(
                "The 'kind' must be 'integer', 'float', or "
                "'string', not '{}'".format(value)
            )
        self._data['kind'] = value

    @property
    def units(self):
        """The units, as a string. These need to be compatible with
        pint"""
        if 'units' not in self._data:
            self._data['units'] = self._data['default_units']

        if self._data['units'] is None:
            return self['default_units']

        return self._data['units']

    @units.setter
    def units(self, value):

        logger.debug("units: value = '{}'".format(value))

        if value == '':
            value = None
        if value is None:
            self.dimensionality = None
        else:
            tmp = ureg(value)
            logger.debug("   tmp = '{}'".format(tmp))
            if self.dimensionality is None:
                self.dimensionality = tmp.dimensionality

            logger.debug(
                "   dimensionality = '{}'".format(self.dimensionality)
            )

            if tmp.dimensionality != self.dimensionality:
                raise RuntimeError(
                    (
                        "Units '{}' have a different dimensionality than "
                        "the parameters: '{}' != '{}'"
                    ).format(value, tmp.dimensionality, self.dimensionality)
                )
        self._data['units'] = value

    @property
    def default_units(self):
        """The default units, as a string. These need to be compatible with
        pint"""
        return self._data['default_units']

    @default_units.setter
    def default_units(self, value):
        if value == '':
            value = None
        if value is None:
            self.dimensionality = None
        else:
            tmp = ureg(value)
            if self.dimensionality is None:
                self.dimensionality = tmp.dimensionality

            if tmp.dimensionality != self.dimensionality:
                raise RuntimeError(
                    (
                        "The default units '{}' have a different "
                        "dimensionality than the parameters: "
                        "'{}' != '{}'"
                    ).format(value, tmp.dimensionality, self.dimensionality)
                )
        self._data['default_units'] = value

    @property
    def enumeration(self):
        """The possible values for an enumerated type."""
        return self._data['enumeration']

    @property
    def format_string(self):
        """The format string for the value"""
        return self._data['format_string']

    @format_string.setter
    def format_string(self, value):
        self._data['format_string'] = value

    @property
    def description(self):
        """Short description of this parameter, preferable just a
        few words"""
        return self._data['description']

    @description.setter
    def description(self, value):
        self._data['description'] = value

    @property
    def help_text(self):
        """A longer description of this parameter that is suitable
        for e.g. help text."""
        return self._data['help_text']

    @help_text.setter
    def help_text(self, value):
        self._data['help_text'] = value

    @property
    def has_units(self):
        """Does this parameter have units associated?"""
        if self.dimensionality is None:
            return False
        if self.dimensionality == '':
            return False
        return True

    @property
    def is_expr(self):
        """Is the current value a variable reference or
        expression?"""
        if isinstance(self.value, str) and len(self.value) > 0:
            return self.value and self.value[0] == '$'
        else:
            return False

    def get(self, context=None, formatted=False, units=True):
        """Return the value evaluated in the given context"""
        if self.is_expr:
            if context is None:
                global root_context
                if root_context is None:
                    raise RuntimeError('No context available')
                result = eval(self.value[1:], root_context)
            else:
                result = eval(self.value[1:], context)
        else:
            result = self.value

        # If it is an enum, just return that.
        if self.enumeration is not None and result in self.enumeration:
            if self.kind == 'boolean':
                return bool(strtobool(result))
            else:
                return result

        # convert to proper type
        if self.kind == 'integer':
            result = int(result)
        elif self.kind == 'float':
            result = float(result)
        elif self.kind == 'boolean':
            if isinstance(result, str):
                result = bool(strtobool(result))
            elif not isinstance(result, bool):
                result = bool(result)
        elif self.kind == 'list' or self.kind == 'periodic table':
            if not isinstance(result, list):
                if (
                    isinstance(result, str) and len(result) > 0 and
                    result[0] != '$'
                ):
                    result = json.loads(result)
            return result
        elif self.kind == 'dictionary':
            if not isinstance(result, dict):
                result = json.loads(result)
            return result

        # format if requested
        if formatted:
            result = self.format_string.format(result)
            if self.units is not None and self.units != '':
                result += ' ' + self.units

        # and run into pint quantity if requested
        if units and self.units is not None and self.units != '':
            result = Q_(result, self.units)

        return result

    def set(self, value):
        """Set the fields based on the type of value given"""
        if self.kind == 'special' or self.kind == 'periodic table':
            self.value = value
        elif isinstance(value, tuple) or isinstance(value, list):
            if len(value) == 1:
                self.value = value[0]
            elif len(value) == 2:
                self.value = value[0]
                self.units = value[1]
            else:
                raise RuntimeError(
                    "Parameter.set expected a sequence of length "
                    "1 or 2, not '{}'".format(len(value))
                )
        else:
            self.value = value

    def reset(self):
        """Reset to an empty state"""
        self._data = {
            'default': None,
            'kind': None,
            'widget': None,
            'default_units': None,
            'enumeration': None,
            'format_string': None,
            'group': '',
            'description': None,
            'help_text': None,
        }
        self.dimensionality = None

    def widget(self, frame, **kwargs):
        """Return a widget for handling the parameter"""
        # Will this keep the graphics isolated?
        import seamm_widgets as sw

        logger.debug('Creating widget for {}'.format(type(self)))

        if self._widget is not None:
            logger.debug('   Destroying existing widget.')
            try:
                self._widget.destroy()
            except Exception:
                pass

        labeltext = kwargs.pop('labeltext', self.description)

        if self.kind == 'special':
            module_name, class_name = self['widget'].split('.')
            mdl = importlib.import_module(module_name)
            cls = getattr(mdl, class_name)
            w = cls(frame, labeltext=labeltext, **kwargs)
            w.set(self.value)
        elif self.kind == 'periodic table':
            w = sw.PeriodicTable(frame, **kwargs)
            w.set(self.value)
        elif self.enumeration is not None:
            if len(self.enumeration) > 0:
                width = max(len(x) for x in self.enumeration)
                if width < 10:
                    width = 10
            else:
                width = 10
            if self.dimensionality is None:
                logger.debug('    making LabeledCombobox')
                w = sw.LabeledCombobox(
                    frame,
                    labeltext=labeltext,
                    values=self.enumeration,
                    width=width,
                    **kwargs
                )
                w.set(self.value)
            else:
                logger.debug('   making UnitCombobox')
                w = sw.UnitCombobox(
                    frame,
                    labeltext=labeltext,
                    values=self.enumeration,
                    width=width,
                    **kwargs
                )
                w.set(self.value, self.units)
        else:
            if self.dimensionality is None:
                logger.debug('   making LabeledEntry')
                w = sw.LabeledEntry(frame, labeltext=labeltext, **kwargs)
                w.set(self.value)
            else:
                logger.debug('   making UnitEntry')
                w = sw.UnitEntry(frame, labeltext=labeltext, **kwargs)
                w.set(self.value, self.units)

        self._widget = w

        logger.debug('   returning {}'.format(w))
        return w

    def set_from_widget(self):
        """Set the value from the widget, ignoring if there is no widget.
        """
        if self._widget is not None:
            self.set(self._widget.get())

    def reset_widget(self):
        """Reset the values in the widget, if it has been created.
        """
        if self._widget is not None:
            if self.dimensionality is None:
                self._widget.set(self.value)
            else:
                self._widget.set(self.value, self.units)

    def to_dict(self):
        """Convert into a string suitable for editing"""
        result = dict()
        # if self['kind'] == 'list':
        #     result['value'] = json.dumps(self.value)
        # elif self['kind'] == 'dict':
        #     result['value'] = json.dumps(self.value)
        # else:
        #     result['value'] = self.value
        result['value'] = self.value
        result['units'] = self.units
        return result

    def update(self, data):
        """Update values from a dict

        This assumes that the static data such as 'kind' and
        'default' has been created already.
        """

        logger.debug('Parameter.update....')
        for key, value in data.items():
            logger.debug('{:>10s} {}'.format(key, value))
            if key in ('value', 'default'):
                # if self['kind'] in ('list', 'dictionary'):
                #     self._data[key] = json.loads(value)
                # else:
                self._data[key] = value
            elif key == 'units':
                self._data[key] = value
            elif key not in self:
                raise RuntimeError(
                    'update: dictionary not compatible with Parameters,'
                    " which do not have an attribute '{}'".format(key)
                )
            else:
                self._data[key] = value

        # Update the dimensionality if needed
        if 'units' in self._data:
            self.units = self._data['units']
        if 'default_units' in self._data:
            self.default_units = self._data['default_units']

    def debug_print(self):
        logger.debug(
            '\nParameter instance:\n{}'.format(pprint.pformat(self._data))
        )


class Parameters(collections.abc.MutableMapping):
    """A dict-like container for parameters"""

    def __init__(self, defaults={}, data=None):
        """Create an instance, optionally from a dict"""

        logger.debug('\nParameters.__init__')
        logger.debug(pprint.pformat(defaults))

        self.defaults = defaults
        logger.debug('\ndefaults:\n{}'.format(pprint.pformat(defaults)))

        self._data = {}

        self.initialize()

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("\nafter defaults:")
            for key, value in self.items():
                logger.debug(
                    '  {}: {}'.format(key, pprint.pformat(value._data))
                )

        if data:
            if isinstance(data, dict):
                self.update(data)

                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("\nafter data:")
                    for key, value in self.items():
                        logger.debug(
                            '  {}: {}'.format(
                                key, pprint.pformat(value._data)
                            )
                        )
            else:
                raise RuntimeError(
                    'A Parameters object can be initialized with a dict object'
                )

    def __getitem__(self, key):
        """Allow [] access to the dictionary!"""
        return self._data[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._data[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        del self._data[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._data)

    def __len__(self):
        """The len() command"""
        return len(self._data)

    def __repr__(self):
        """The string representation of this object"""
        return repr(self._data)

    def __str__(self):
        """The pretty string representation of this object"""
        return pprint.pformat(self.to_dict())

    def __contains__(self, item):
        """Return a boolean indicating if a key exists."""
        if item in self._data:
            return True
        return False

    def __eq__(self, other):
        """Return a boolean if this object is equal to another"""
        return self._data == other._data

    def copy(self):
        """Return a shallow copy of the dictionary"""
        return self._data.copy()

    def to_dict(self):
        """Return a new dictionary with the pertinent data

        The Parameter class only saves the value and units,
        as everything else comes form the constructor below
        """
        data = {}
        for key in self:
            try:
                data[key] = self[key].to_dict()
            except:  # noqa: E722
                logger.critical(
                    (
                        "An error occurred in Parameters.to_dict "
                        "with key '{}'"
                    ).format(key)
                )
                logger.critical(
                    ("The type of the key is '{}'").format(type(self[key]))
                )
                raise
        return data

    def from_dict(self, data):
        """Recreate the object from a dictionary"""
        self._data = dict()
        # Put back in all the constant data
        self.initialize()
        # and update with the new data
        self.update(data)

    def initialize(self):
        for key, value in self.defaults.items():
            self[key] = Parameter(value)

    def update(self, data):
        for key in data:
            self[key].update(data[key])

    def values_to_dict(self):
        """Return a dict of the raw values of the parameters
        formatted for printing"""

        data = {}
        for key in self:
            try:
                data[key] = str(self[key])
            except Exception as e:
                logger.warning("Cannot format '{}': {}".format(key, str(e)))
                data[key] = '#err#'

        return data

    def current_values_to_dict(
        self, context=None, formatted=False, units=True
    ):
        """Return the current values of the parameters, resolving
        any expressions, etc. in the given context or the root
        context is none is given."""

        data = {}
        for key in self:
            data[key] = self[key].get(
                context=context, formatted=formatted, units=units
            )

        return data

    def set_from_widgets(self):
        """Convenience function to set the parameters from their widgets."""
        for key in self:
            self[key].set_from_widget()

    def reset_widgets(self):
        """Convenience function to reset the widgets to the current value."""
        for key in self:
            try:
                self[key].reset_widget()
            except ValueError as e:
                logger.warning(
                    'Error resetting widget for {}: {}'.format(key, str(e))
                )
                raise
            except Exception:
                raise



# -*- coding: utf-8 -*-

"""Run a SEAMM flowchart.

SEAMM flowcharts have a 'magic' line, so that they can be executed directly.
Or, run_flowchart can be invoked with the name of flowchart.
"""

import configargparse
import cpuinfo
import datetime
import fasteners
import json
import locale
import logging
import seamm
import seamm_util.printing as printing
import os
import os.path
import re
import shutil
import sys
import time

logger = logging.getLogger(__name__)
variables = seamm.Variables()


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def run(job_id=None, wdir=None, setup_logging=True):
    """The standalone flowchart app
    """

    parser = configargparse.ArgParser(
        auto_env_var_prefix='',
        default_config_files=[
            '/etc/seamm/seamm.ini',
            '~/.seamm/seamm.ini',
        ],
        description='Execute a SEAMM flowchart'
    )

    parser.add_argument(
        "--standalone",
        action='store_true',
        help="Run this workflow as-is without using the datastore, etc."
    )
    parser.add_argument(
        '--seamm-configfile',
        is_config_file=True,
        default=None,
        help='a configuration file to override others'
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose_count",
        action="count",
        default=0,
        help="increases log verbosity for each occurence."
    )
    parser.add_argument(
        "--title",
        dest="title",
        default='',
        action="store",
        env_var='SEAMM_TITLE',
        help="The title for this run."
    )
    parser.add_argument(
        "--datastore",
        dest="datastore",
        default='.',
        action="store",
        env_var='SEAMM_DATASTORE',
        help="The datastore (directory) for this run."
    )
    parser.add_argument(
        "--job-id-file",
        dest="job_id_file",
        default=None,
        action="store",
        help="The job_id file to use."
    )
    parser.add_argument(
        "--project",
        dest="projects",
        action="append",
        env_var='SEAMM_PROJECT',
        help="The project(s) for this job."
    )
    parser.add_argument("--force", dest="force", action='store_true')
    parser.add_argument(
        "--output",
        choices=['files', 'stdout', 'both'],
        default='files',
        help='whether to put the output in files, direct to stdout, or both'
    )
    parser.add_argument(
        "filename",
        nargs='?',
        help='the filename of the flowchart'
    )  # yapf: disable

    args, unknown = parser.parse_known_args()

    # Whether to just run as-is, without getting a job_id, using the
    # datastore, etc.
    standalone = 'standalone' in args and args.standalone

    if setup_logging:
        # Set up logging level to WARNING by default, going more verbose
        # for each new -v, to INFO and then DEBUG and finally ALL with 3 -v's

        numeric_level = max(3 - args.verbose_count, 0) * 10
        logging.basicConfig(level=numeric_level)

    # Create the working directory where files, output, etc. go.
    # At the moment this is datastore/job_id

    if standalone:
        print('Running in standalone mode.')
        wdir = os.getcwd()
    else:
        datastore = os.path.expanduser(args.datastore)

        if job_id is None:
            if args.job_id_file is None:
                job_id_file = os.path.join(datastore, 'job.id')

            # Get the job_id from the file, creating the file if necessary
            job_id = get_job_id(job_id_file)
        if wdir is None:
            if args.projects is None:
                projects = ['default']
            else:
                projects = args.projects

            # And put it all together
            wdir = os.path.abspath(
                os.path.join(
                    datastore, 'projects', projects[0],
                    'Job_{:06d}'.format(job_id)
                )
            )

            if os.path.exists(wdir):
                if args.force:
                    shutil.rmtree(wdir)
                else:
                    msg = "Directory '{}' exists, use --force to overwrite"\
                          .format(wdir)

                    logging.critical(msg)
                    sys.exit(msg)

            os.makedirs(wdir, exist_ok=False)

    logging.info("The working directory is '{}'".format(wdir))

    # Set up the root printer, and add handlers to print to the file
    # 'job.out' in the working directory and to stdout, as requested
    # in the options. Since all printers are children of the root
    # printer, all output at the right levels will flow here

    printer = printing.getPrinter()

    # Set up our formatter
    formatter = logging.Formatter(fmt='{message:s}', style='{')

    # A handler for stdout
    if standalone or wdir is None:
        console_handler = logging.StreamHandler()
        # console_handler.setLevel(printing.JOB)
        console_handler.setLevel(printing.NORMAL)
        console_handler.setFormatter(formatter)
        printer.addHandler(console_handler)

    # A handler for the file
    file_handler = logging.FileHandler(os.path.join(wdir, 'job.out'))
    file_handler.setLevel(printing.NORMAL)
    file_handler.setFormatter(formatter)
    printer.addHandler(file_handler)

    # And ... finally ... run!
    printer.job("Running in directory '{}'".format(wdir))

    flowchart_path = os.path.join(wdir, 'flowchart.flow')
    if args.filename is not None:
        # copy the flowchart to the root directory for later reference
        shutil.copy2(args.filename, flowchart_path)

    flowchart = seamm.Flowchart(directory=wdir, output=args.output)
    flowchart.read(flowchart_path)

    # Change to the working directory and run the flowchart
    with cd(wdir):
        if os.path.exists('job_data.json'):
            with open('job_data.json', 'r') as fd:
                data = json.load(fd)
        else:
            data = {}

        # Set up the initial metadata for the job.
        data.update(cpuinfo.get_cpu_info())
        if 'command line' not in data:
            data['command line'] = sys.argv
        if 'title' not in data:
            data['title'] = args.title
        data['working directory'] = wdir
        data['start time'] = time.strftime("%Y-%m-%d %H:%M:%S %Z")
        data['state'] = 'started'
        if not standalone:
            if 'projects' not in data:
                data['projects'] = projects
            data['datastore'] = datastore
            data['job id'] = job_id

        # Output the initial metadate for the job.
        with open('job_data.json', 'w') as fd:
            json.dump(data, fd, indent=3, sort_keys=True)

        t0 = time.time()
        pt0 = time.process_time()

        # And run the flowchart
        try:
            exec = seamm.ExecFlowchart(flowchart)
            exec.run(root=wdir)
            data['state'] = 'finished'
        except Exception as e:
            data['state'] = 'error'
            data['error type'] = type(e).__name__
            data['error message'] = str(e)

        # Wrap things up
        t1 = time.time()
        pt1 = time.process_time()
        data['end time'] = time.strftime("%Y-%m-%d %H:%M:%S %Z")
        t = t1 - t0
        pt = pt1 - pt0
        data['elapsed time'] = t
        data['process time'] = pt

        with open('job_data.json', 'w') as fd:
            json.dump(data, fd, indent=3, sort_keys=True)

        printer.job(
            "\nProcess time: {} ({:.3f} s)".format(
                datetime.timedelta(seconds=pt), pt
            )
        )
        printer.job(
            "Elapsed time: {} ({:.3f} s)".format(
                datetime.timedelta(seconds=t), t
            )
        )


def get_job_id(filename):
    """Get the next job id from the given file.

    This uses the fasteners module to provide locking so that
    only one job at a time can access the file, so that the job
    ids are unique and monotonically increasing.
    """

    lock_file = filename + '.lock'
    lock = fasteners.InterProcessLock(lock_file)
    locked = lock.acquire(blocking=True, timeout=5)

    if locked:
        if not os.path.isfile(filename):
            job_id = 1
            with open(filename, 'w') as fd:
                fd.write('!MolSSI job_id 1.0\n')
                fd.write('1\n')
            lock.release()
        else:
            with open(filename, 'r+') as fd:
                line = fd.readline()
                pos = fd.tell()
                if line == '':
                    lock.release()
                    raise EOFError(
                        "job_id file '{}' is empty".format(filename)
                    )
                line = line.strip()
                match = re.fullmatch(
                    r'!MolSSI job_id ([0-9]+(?:\.[0-9]+)*)', line
                )
                if match is None:
                    lock.release()
                    raise RuntimeError(
                        'The job_id file has an incorrect header: {}'
                        .format(line)
                    )
                line = fd.readline()
                if line == '':
                    lock.release()
                    raise EOFError(
                        "job_id file '{}' is truncated".format(filename)
                    )
                try:
                    job_id = int(line)
                except TypeError:
                    raise TypeError(
                        "The job_id in file '{}' is not an integer: {}".format(
                            filename, line
                        )
                    )
                job_id += 1
                fd.seek(pos)
                fd.write('{:d}\n'.format(job_id))
    else:
        raise RuntimeError(
            "Could not lock the job_id file '{}'".format(filename)
        )

    return job_id


if __name__ == "__main__":
    locale.setlocale(locale.LC_ALL, '')
    run()



# -*- coding: utf-8 -*-

"""A node in a flowchart


"""

import bibtexparser
import collections.abc
import jinja2
import json
import logging
import pkg_resources
import reference_handler
import seamm
import seamm_util  # MUST come after seamm
from seamm_util.printing import FormattedText as __
import seamm_util.printing as printing
import numpy as np
import os.path
import pandas
import uuid

logger = logging.getLogger(__name__)
job = printing.getPrinter()


class Node(collections.abc.Hashable):

    def __init__(self, flowchart=None, title='', extension=None, module=None):
        """Initialize a node

        Keyword arguments:
        """

        self._uuid = uuid.uuid4().int
        self.module = module
        self.parent = None
        self.flowchart = flowchart
        self._title = title
        self._description = ''
        self._id = None
        self.extension = extension
        self._visited = False
        self._references = None
        self._jinja_env = None
        self._graphs = None

        self.parameters = None  # Object containing control parameters

        self.x = None
        self.y = None
        self.w = None
        self.h = None

        # Set up our formatter for printing
        self.formatter = logging.Formatter(fmt='{message:s}', style='{')

        # Setup the bibliography
        self.bibliography = {}
        if self.module:
            filepath = pkg_resources.resource_filename(
                self.module, 'data/references.bib'
            )
            logger.info("bibliography file path = '{}'".format(filepath))

            if os.path.exists(filepath):
                with open(filepath) as fd:
                    tmp = bibtexparser.load(fd).entries_dict
                writer = bibtexparser.bwriter.BibTexWriter()
                for key, data in tmp.items():
                    self.bibliography[key] = writer._entry_to_bibtex(data)

    def __hash__(self):
        """Make iterable!"""
        return self._uuid

    def __eq__(self, other):
        return (
            self.__class__ == other.__class__ and
            self.__hash__() == other.__hash__()
        )

    @property
    def uuid(self):
        """The uuid of the node"""
        return self._uuid

    @property
    def title(self):
        """The title to display"""
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def tag(self):
        """The string representation of the uuid of the node"""
        return 'node=' + str(self._uuid)

    @property
    def directory(self):
        """Return the directory we should write output to"""
        return os.path.join(self.flowchart.root_directory, *self._id)

    @property
    def visited(self):
        """Whether this node has been visited in a traversal"""
        return self._visited

    @visited.setter
    def visited(self, value):
        self._visited = bool(value)

    @property
    def description(self):
        """A textual description of this node"""
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def indent(self):
        length = len(self._id)
        if length <= 1:
            return ''
        if length > 2:
            result = (length - 2) * (3 * ' ' + '.') + 4 * ' '
        else:
            result = 4 * ' '
        return result

    @property
    def header(self):
        """A printable header for this section of output"""
        return (
            'Step {}: {}  {}'.format(
                '.'.join(str(e) for e in self._id), self.title, self.version
            )
        )

    @property
    def references(self):
        """The reference handle for citations."""
        if self._references is None:
            filename = os.path.join(
                self.flowchart.root_directory, 'references.db'
            )
            self._references = reference_handler.Reference_Handler(filename)

        return self._references

    @references.setter
    def references(self, value):
        if self._references is not None:
            self._references.conn.commit()
            self._references.__del__()

        self._references = value

    def set_uuid(self):
        self._uuid = uuid.uuid4().int

        # Need to correct all edges to other nodes
        raise NotImplementedError('set_uuid not implemented yet!')

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        if self.visited:
            return None
        else:
            self.visited = True
            self._id = node_id
            return self.next()

    def reset_id(self):
        """Reset the id for node"""
        self._id = None

    def get_gui_data(self, key, gui=None):
        """Return an element from the GUI dictionary"""
        if gui is None:
            return self._gui_data[seamm.Flowchart.graphics][key]
        else:
            return self._gui_data[gui][key]

    def set_gui_data(self, key, value, gui=None):
        """Set an element of the GUI dictionary"""
        if gui is None:
            if seamm.Flowchart.graphics not in self._gui_data:
                self._gui_data[seamm.Flowchart.graphics] = {}
            self._gui_data[seamm.Flowchart.graphics][key] = value
        else:
            if gui not in self._gui_data:
                self._gui_data[gui] = {}
            self._gui_data[gui][key] = value

    def connections(self):
        """Return a list of all the incoming and outgoing edges
        for this node, giving the anchor points and other node
        """

        result = self.flowchart.edges(self)
        return result

    def remove_edge(self, edge):
        """Remove a given edge, or all edges if 'all' is given
        """

        if isinstance(edge, str) and edge == 'all':
            for direction, obj in self.connections():
                self.remove_edge(obj)
        else:
            self.flowchart.graph.remove_edge(
                edge.node1, edge.node2, edge.edge_type, edge.edge_subtype
            )

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """
        return (
            'This node has no specific description. '
            "Override the method 'description_text' "
            'to provide the description.'
        )

    def describe(self):
        """Write out information about what this node will do
        """

        self.visited = True

        # The description
        job.job(__(self.description_text(), indent=self.indent))

        next_node = self.next()

        if next_node is None or next_node.visited:
            return None
        else:
            return next_node

    def run(self, printer=None):
        """Do whatever we need to do! The base class does nothing except
        return the next node.
        """

        # Create the directory for writing output and files
        os.makedirs(self.directory, exist_ok=True)

        if printer is not None:
            # Setup up the printing for this step
            self.setup_printing(printer)

        next_node = self.next()
        if next_node:
            logger.debug('returning next_node: {}'.format(next_node))
        else:
            logger.debug('returning next_node: None')

        return next_node

    def next(self):
        """Return the next node in the flow
        """

        for edge in self.flowchart.edges(self, direction='out'):
            if edge.edge_subtype == 'next':
                logger.debug('Next node is: {}'.format(edge.node2))
                return edge.node2

        logger.debug('Reached the end of the flowchart')
        return None

    def previous(self):
        """Return the previous node in the flow
        """

        for edge in self.flowchart.edges(self, direction='in'):
            if edge.edge_type == 'execution' and \
               edge.edge_subtype == 'next':
                return edge.node1

        return None

    def get_input(self):
        """Return the input from this subnode, usually used for
        building up the input for the executable."""

        return ''

    def to_json(self):
        return json.dumps(self.to_dict(), cls=seamm_util.JSONEncoder)

    def to_dict(self):
        """serialize this object and everything it contains as a dict"""
        data = {
            'item': 'object',
            'module': self.__module__,
            'class': self.__class__.__name__,
            'extension': self.extension
        }
        data['attributes'] = {}
        for key in self.__dict__:
            if key == 'flowchart':
                continue
            if key == 'parent':
                continue
            if key == 'formatter':
                continue
            if key == 'parser':
                continue
            if key == 'options':
                continue
            if key == 'unknown':
                continue
            if 'flowchart' in key:
                # Have a subflowchart!
                data[key] = self.__dict__[key].to_dict()
            else:
                data['attributes'][key] = self.__dict__[key]
        return data

    def from_dict(self, data):
        """un-serialize object and everything it contains from a dict"""
        if data['item'] != 'object':
            raise RuntimeError('The data for restoring the object is invalid')
        if data['class'] != self.__class__.__name__:
            raise RuntimeError(
                'Trying to restore a {}'.format(self.__class__.__name__) +
                ' from data for a {}'.format(data['class'])
            )
        for key in data:
            if key == 'attributes':
                attributes = data['attributes']
                for subkey in attributes:
                    self.__dict__[subkey] = attributes[subkey]
            elif 'flowchart' in key:
                self.__dict__[key].from_dict(data[key])

    def default_edge_subtype(self):
        """Return the default subtype of the edge. Usually this is 'next'
        but for nodes with two or more edges leaving them, such as a loop, this
        method will return an appropriate default for the current edge. For
        example, by default the first edge emanating from a loop-node is the
        'loop' edge; the second, the 'exit' edge.

        A return value of 'too many' indicates that the node exceeds the number
        of allowed exit edges.
        """

        # how many outgoing edges are there?
        n_edges = len(self.flowchart.edges(self, direction='out'))

        logger.debug('node.default_edge_subtype, n_edges = {}'.format(n_edges))

        if n_edges == 0:
            return ""
        else:
            return "too many"

    def analyze(self, indent='', **kwargs):
        """Analyze the output of the calculation
        """
        return

    def get_value(self, variable_or_value):
        """Return the value of the workspace variable is <variable_or_value>
        is the name of a variable. Otherwise, simply return the value of
        <variable_or_value>.

        This provides a convenient way to handle both values and variables
        in widgets. A reference to a variable is $<name> or ${name}, and is
        replaced by the contents of the variable. If the text is not a
        reference to a variable then the value passed in is returned
        unchanged.
        """

        return seamm.flowchart_variables.value(variable_or_value)

    def get_variable(self, variable):
        """Get the value of a variable, which must exist
        """

        return seamm.flowchart_variables.get_variable(variable)

    def set_variable(self, variable, value):
        """Set the value of a variable in the workspace. The name of the
        variable maybe a plain string, or be $<name> or ${<name>}
        """

        seamm.flowchart_variables.set_variable(variable, value)

    def variable_exists(self, variable):
        """Return whether a varable exists in the workspace
        """

        return seamm.flowchart_variables.exists(variable)

    def delete_variable(self, variable):
        """Delete a variable in the workspace
        """

        seamm.flowchart_variables.delete(variable)

    def setup_printing(self, printer):
        """Establish the handlers for printing as controlled by
        options
        """

        # Control output going to the main job printer
        # If we are in a loop, don't print to the job output, except
        # at the JOB level
        # job.setLevel(printing.NORMAL)
        # for segment in self._id:
        #     if str(segment)[0:5] == 'iter_':
        #         job.setLevel(printing.JOB)
        #         break

        # First remove an existing handlers
        self.close_printing(printer)

        # A handler for stdout
        console_handler = logging.StreamHandler()
        console_handler.setLevel(printing.JOB)
        console_handler.setFormatter(self.formatter)
        printer.addHandler(console_handler)

        # A handler for the file
        file_handler = logging.FileHandler(
            os.path.join(self.directory, 'step.out'), delay=True
        )
        file_handler.setLevel(printing.NORMAL)
        file_handler.setFormatter(self.formatter)
        printer.addHandler(file_handler)

        # # A handler for the job file
        # wdir = self.flowchart.root_directory
        # job_file_handler = logging.FileHandler(os.path.join(wdir, 'job.out'))
        # # job_file_handler.setLevel(printing.JOB)
        # job_file_handler.setLevel(printing.NORMAL)
        # job_file_handler.setFormatter(self.formatter)
        # printer.addHandler(job_file_handler)

    def close_printing(self, printer):
        """Close the handlers for printing, so that buffers are
        flushed, files closed, etc.
        """
        for handler in printer.handlers:
            printer.removeHandler(handler)

    def job_output(self, text):
        """Temporary!"""
        job.job(text)

    def store_results(
        self, data={}, properties=None, results=None, create_tables=True
    ):
        """Store results in variables and tables, as requested

        Keywords:

        properties (dict): a dictionary of properties
        results (dict): a dictionary of results from the calculation
        create_tables (bool): whether to create tables as needed

        Each item in 'results' is itself a dictionary. If the following keys
        are in the dictionary, the appropriate action is taken:

        'variable' -- is the name of a variable to store the result in
        'table' -- the name of the table, and
        'column' -- is the column name for the result in the table.
        """

        for key, value in results.items():
            # Check for storing in a variable
            if 'variable' in value:
                variable = self.get_value(value['variable'])
                self.set_variable(variable, data[key])

            # and table
            if 'table' in value:
                tablename = value['table']
                column = self.get_value(value['column'])
                # Does the table exist?
                if not self.variable_exists(tablename):
                    if create_tables:
                        table = pandas.DataFrame()
                        self.set_variable(
                            tablename, {
                                'type': 'pandas',
                                'table': table,
                                'defaults': {},
                                'loop index': False,
                                'current index': 0
                            }
                        )
                    else:
                        raise RuntimeError(
                            "Table '{}' does not exist.".format(tablename)
                        )

                table_handle = self.get_variable(tablename)
                table = table_handle['table']

                # create the column as needed
                if column not in table.columns:
                    kind = properties[key]['type']
                    if kind == 'boolean':
                        default = False
                    elif kind == 'integer':
                        default = 0
                    elif kind == 'float':
                        default = np.nan
                    else:
                        default = ''

                    table_handle['defaults'][column] = default
                    table[column] = default

                # and put the value in (finally!!!)
                row_index = table_handle['current index']
                if key in data:
                    table.at[row_index, column] = data[key]

    def create_figure(
        self, title='', template='line.graph_template', module_path=None
    ):
        """Create a new figure.

        Parameters
        ----------
        title : str, optional
        template : str, optional
            The Jinja template for the desired graph. Defaults to
            'line.graph_template'

        Returns
        -------
        seamm_util.Figure
        """

        if self._jinja_env is None:
            # The order of the loaders is important! They are searched
            # in order, so the first has precedence. This searches the
            # current package first, then looks in the main SEAMM
            # templates.
            if module_path is None:
                logger.info("Reading graph templates from 'seamm'")
                loaders = [jinja2.PackageLoader('seamm')]
            else:
                logger.info(
                    "Reading graph templates from the following modules, "
                    "in order"
                )
                loaders = []
                for module in module_path:
                    logger.info('\t' + module)
                    loaders.append(jinja2.PackageLoader(module))

            self._jinja_env = jinja2.Environment(
                loader=jinja2.ChoiceLoader(loaders)
            )

        figure = seamm_util.Figure(
            jinja_env=self._jinja_env, template=template, title=title
        )
        return figure
